"""Inspiration from the deskbar pidgin plugin and from the gajim kupfer
plugin"""

__kupfer_name__ = _("Pidgin")
__kupfer_sources__ = ("ContactsSource",)
__kupfer_actions__ = ("OpenChat", "SendMessage")
__description__ = _("Access to Pidgin Contacts")
__author__ = (
    "Chmouel Boudjnah <chmouel@chmouel.com>, "
    "Ulrik Sverdrup <ulrik.sverdrup@gmail.com>"
)
__version__ = "2017.1"

import typing as ty

import dbus

from kupfer import icons, plugin_support
from kupfer.obj import Action, Leaf, TextLeaf, TextSource
from kupfer.obj.apps import AppLeafContentMixin
from kupfer.obj.contacts import EMAIL_KEY, NAME_KEY, ContactLeaf
from kupfer.obj.grouping import ToplevelGroupingSource
from kupfer.support import pretty, scheduler, weaklib
from kupfer.support.validators import is_valid_email

if ty.TYPE_CHECKING:
    from gettext import gettext as _
    from gettext import ngettext

__kupfer_settings__ = plugin_support.PluginSettings(
    {
        "key": "show_offline",
        "label": _("Show offline contacts"),
        "type": bool,
        "value": False,
    },
)


plugin_support.check_dbus_connection()

# Contact data constants
_PIDGIN_ACCOUNT = "_PIDGIN_ACCOUNT"
_PIDGIN_JID = "_PIDGIN_JID"

# D-Bus "addresses"
_SERVICE_NAME = "im.pidgin.purple.PurpleService"
_OBJECT_NAME = "/im/pidgin/purple/PurpleObject"
_IFACE_NAME = "im.pidgin.purple.PurpleInterface"


def _create_dbus_connection(activate=False):
    """Create dbus connection to Pidgin
    @activate: true=starts pidgin if not running
    """
    interface = None
    obj = None
    sbus = dbus.SessionBus()

    try:
        # check for running pidgin (code from note.py)
        proxy_obj = sbus.get_object(
            "org.freedesktop.DBus", "/org/freedesktop/DBus"
        )
        dbus_iface = dbus.Interface(proxy_obj, "org.freedesktop.DBus")
        if activate or dbus_iface.NameHasOwner(_SERVICE_NAME):
            obj = sbus.get_object(_SERVICE_NAME, _OBJECT_NAME)

        if obj:
            interface = dbus.Interface(obj, _IFACE_NAME)

    except dbus.exceptions.DBusException as err:
        pretty.print_debug(err)

    return interface


def _send_message_to_contact(pcontact, message, present=False):
    """Send @message to PidginContact @pcontact"""
    interface = _create_dbus_connection()
    if not interface:
        return

    account = pcontact[_PIDGIN_ACCOUNT]
    jid = pcontact[_PIDGIN_JID]
    conversation = interface.PurpleConversationNew(1, account, jid)
    imiface = interface.PurpleConvIm(conversation)
    interface.PurpleConvImSend(imiface, message)
    if present:
        interface.PurpleConversationPresent(conversation)


class ContactAction(Action):
    def get_required_slots(self) -> tuple[str]:
        raise NotImplementedError

    def item_types(self):
        yield ContactLeaf

    def valid_for_item(self, leaf: Leaf) -> bool:
        leaf = ty.cast("ContactLeaf", leaf)
        return all(slot in leaf for slot in self.get_required_slots())


class OpenChat(ContactAction):
    """Open Chat Conversation Window with jid"""

    # consider it as main action for pidgin contacts
    rank_adjust = 5

    def __init__(self):
        super().__init__(_("Open Chat"))

    def activate(self, leaf, iobj=None, ctx=None):
        _send_message_to_contact(leaf, "", present=True)

    def get_required_slots(self):
        return (_PIDGIN_ACCOUNT, _PIDGIN_JID)


class ChatTextSource(TextSource):
    def get_rank(self):
        return 100

    def get_text_items(self, text):
        textlen = len(text)
        summary = text[:10] + (text[10:11] and "..")
        desc_template = ngettext(
            "%s (%d character)", "%s (%d characters)", textlen
        )
        yield TextLeaf(text, desc_template % (summary, textlen))

    def get_items(self, text):
        return self.get_text_items(text)


class SendMessage(ContactAction):
    """Send chat message directly from Kupfer"""

    def __init__(self):
        super().__init__(_("Send Message..."))

    def activate(self, leaf, iobj=None, ctx=None):
        assert iobj
        _send_message_to_contact(leaf, iobj.object)

    def get_required_slots(self):
        return (_PIDGIN_ACCOUNT, _PIDGIN_JID)

    def requires_object(self):
        return True

    def object_types(self):
        yield TextLeaf

    def object_source(self, for_item=None):
        return TextSource()

    def valid_object(self, iobj, for_item=None):
        # ugly, but we don't want derived text
        return type(iobj) == TextLeaf  # pylint: disable=unidiomatic-typecheck # noqa: E721


class PidginContact(ContactLeaf):
    """Leaf represent single contact from Pidgin"""

    grouping_slots = (*ContactLeaf.grouping_slots, EMAIL_KEY)

    def __init__(
        self, jid, name, account, icon, protocol, available, status_message
    ):
        slots = {
            EMAIL_KEY: jid,
            NAME_KEY: name or jid,
            _PIDGIN_ACCOUNT: account,
            _PIDGIN_JID: jid,
        }
        if not is_valid_email(jid):
            slots[EMAIL_KEY] = None

        ContactLeaf.__init__(self, slots, name or jid)

        self.kupfer_add_alias(jid)

        self._description = _("[%(status)s] %(userid)s/%(service)s") % {
            "status": _("Available") if available else _("Away"),
            "userid": jid,
            "service": protocol,
        }

        if status_message:
            self._description += f"\n{status_message}"

        self.account = account
        self.jid = jid
        self.icon = icon
        self.protocol = protocol

    def repr_key(self):
        # the repr key should be persistent and hopefully unique
        return f"{self.protocol}, {self.object[_PIDGIN_JID]}"

    def get_description(self):
        return self._description

    def get_thumbnail(self, width, height):
        if not self.icon:
            return None

        return icons.get_pixbuf_from_file(self.icon, width, height)

    def get_icon_name(self):
        return "stock_person"


class ContactsSource(AppLeafContentMixin, ToplevelGroupingSource):
    """Get contacts from all on-line accounts in Pidgin via DBus"""

    appleaf_content_id = "pidgin"
    source_use_cache = False

    def __init__(self):
        ToplevelGroupingSource.__init__(self, _("Pidgin Contacts"), "Contacts")
        self._version = 5
        self.all_buddies = {}
        self._buddy_update_timer = None
        self._buddy_update_queue = None

    def initialize(self):
        ToplevelGroupingSource.initialize(self)
        self._install_dbus_signal()
        self._buddy_update_timer = scheduler.Timer()
        self._buddy_update_queue = set()

    def _get_pidgin_contact(
        self, interface, buddy, account=None, protocol=None
    ):
        if not account:
            account = interface.PurpleBuddyGetAccount(buddy)

        if not protocol:
            protocol = interface.PurpleAccountGetProtocolName(account)

        jid = interface.PurpleBuddyGetName(buddy)
        name = interface.PurpleBuddyGetAlias(buddy)
        icon = None
        if (_icon := interface.PurpleBuddyGetIcon(buddy)) != 0:
            icon = interface.PurpleBuddyIconGetFullPath(_icon)

        presenceid = interface.PurpleBuddyGetPresence(buddy)
        statusid = interface.PurplePresenceGetActiveStatus(presenceid)
        availability = interface.PurplePresenceIsAvailable(presenceid)
        status_message = interface.PurpleStatusGetAttrString(
            statusid, "message"
        )

        return PidginContact(
            jid, name, account, icon, protocol, availability, status_message
        )

    def _get_all_buddies(self):
        interface = _create_dbus_connection()
        if interface is None:
            return

        accounts = interface.PurpleAccountsGetAllActive()
        show_offline = __kupfer_settings__["show_offline"]
        for account in accounts:
            buddies = interface.PurpleFindBuddies(account, dbus.String(""))
            protocol = interface.PurpleAccountGetProtocolName(account)

            for buddy in buddies:
                if not (show_offline or interface.PurpleBuddyIsOnline(buddy)):
                    continue

                self.all_buddies[buddy] = self._get_pidgin_contact(
                    interface, buddy, protocol=protocol, account=account
                )

    def _remove_buddies_not_connected(self):
        """Remove buddies that belong to accounts no longer connected"""
        if not self.all_buddies:
            return

        interface = _create_dbus_connection()
        if interface is None:
            self.all_buddies = {}
            return

        try:
            # extra careful as this will fail when Pidgin is Quitting
            accounts = interface.PurpleAccountsGetAllActive()
        except dbus.DBusException:
            self.all_buddies = {}
            return

        is_disconnected = interface.PurpleAccountIsDisconnected
        conn_accounts = {a for a in accounts if not is_disconnected(a)}
        for buddy, pcontact in list(self.all_buddies.items()):
            if pcontact.account not in conn_accounts:
                del self.all_buddies[buddy]

    def _signing_off(self, conn):
        self.output_debug("Pidgin Signing Off", conn)
        self._remove_buddies_not_connected()
        self.mark_for_update()

    def _update_pending(self):
        """Update all buddies in the update queue"""
        assert self._buddy_update_queue
        interface = _create_dbus_connection()
        if interface is None:
            self._buddy_update_queue.clear()
            return

        show_offline = __kupfer_settings__["show_offline"]
        for buddy in self._buddy_update_queue:
            if show_offline or interface.PurpleBuddyIsOnline(buddy):
                self.output_debug("updating buddy", buddy)
                pcontact = self._get_pidgin_contact(interface, buddy)
                self.all_buddies[buddy] = pcontact
            else:
                self.all_buddies.pop(buddy, None)

        self._buddy_update_queue.clear()
        self.mark_for_update()

    def _buddy_needs_update(self, buddy):
        """add @buddy to the update queue"""
        assert self._buddy_update_timer
        if self._buddy_update_queue is not None:
            self._buddy_update_queue.add(buddy)
            self._buddy_update_timer.set(1, self._update_pending)

    def _buddy_signed_on(self, buddy):
        if buddy not in self.all_buddies:
            self._buddy_needs_update(buddy)

    def _buddy_signed_off(self, buddy):
        if buddy in self.all_buddies:
            del self.all_buddies[buddy]
            self.mark_for_update()

    def _buddy_status_changed(self, buddy, old, new):
        """Callback when status is changed reload the entry
        which get the new status"""
        self._buddy_needs_update(buddy)

    def _install_dbus_signal(self):
        """Add signals to pidgin when buddy goes offline or
        online to update the list"""
        try:
            session_bus = dbus.Bus()
        except dbus.DBusException:
            return

        weaklib.dbus_signal_connect_weakly(
            session_bus,
            "SigningOff",
            self._signing_off,
            dbus_interface=_IFACE_NAME,
        )

        weaklib.dbus_signal_connect_weakly(
            session_bus,
            "BuddySignedOn",
            self._buddy_signed_on,
            dbus_interface=_IFACE_NAME,
        )

        weaklib.dbus_signal_connect_weakly(
            session_bus,
            "BuddyStatusChanged",
            self._buddy_status_changed,
            dbus_interface=_IFACE_NAME,
        )

        weaklib.dbus_signal_connect_weakly(
            session_bus,
            "BuddySignedOff",
            self._buddy_signed_off,
            dbus_interface=_IFACE_NAME,
        )

    def get_items(self):
        if not self.all_buddies:
            self._get_all_buddies()

        return list(self.all_buddies.values())

    def should_sort_lexically(self):
        return True

    def get_icon_name(self):
        return "pidgin"

    def provides(self):
        yield PidginContact



"""
This is a Zeal search plugin.
"""

from __future__ import annotations

__kupfer_name__ = _("Zeal Search")
__kupfer_sources__ = ("ZealDocsetsSource",)
__kupfer_actions__ = ("ZealSearch", "ZealSearchInDocset", "ZealSearchFor")
__description__ = _(
    "Search in Zeal, offline documentation browser for software developers."
)
__version__ = "1.1"
__author__ = "Karol Będkowski"

import json
import os
import typing as ty
from pathlib import Path

from kupfer import icons, launch, plugin_support
from kupfer.obj import Action, Leaf, Source, TextLeaf
from kupfer.obj.apps import AppLeafContentMixin
from kupfer.obj.helplib import FilesystemWatchMixin, NonpersistentToken

if ty.TYPE_CHECKING:
    from gettext import gettext as _

plugin_support.check_command_available("zeal")


class ZealSearch(Action):
    def __init__(self):
        Action.__init__(self, _("Zeal Search"))

    def activate(self, leaf, iobj=None, ctx=None):
        launch.spawn_async(["zeal", leaf.object])

    def get_description(self):
        return _("Search in Zeal")

    def get_gicon(self):
        return icons.ComposedIcon("zeal", "edit-find")

    def get_icon_name(self):
        return "zeal"

    def item_types(self):
        yield TextLeaf


class ZealDocsetsSource(AppLeafContentMixin, Source, FilesystemWatchMixin):
    appleaf_content_id = ("zeal", "org.zealdocs.zeal")
    source_scan_interval: int = 3600

    def __init__(self):
        self.docsets_home = os.path.expanduser(
            "~/.local/share/Zeal/Zeal/docsets"
        )
        super().__init__(_("Zeal Docsets"))
        self.monitor_token = None

    def initialize(self):
        self.monitor_token = self.monitor_directories(self.docsets_home)

    def monitor_include_file(self, gfile):
        return gfile and gfile.get_basename().endswith(".docset")

    def get_items(self):
        docsets_home = Path(self.docsets_home)
        if not docsets_home.is_dir():
            return

        with os.scandir(docsets_home) as docdirs:
            for docdir in docdirs:
                if not docdir.is_dir() or not docdir.name.endswith(".docset"):
                    continue

                meta_file = Path(docdir.path, "meta.json")
                if not meta_file.is_file():
                    continue

                try:
                    with meta_file.open("r", encoding="UTF-8") as meta:
                        content = json.load(meta)
                except OSError:
                    continue

                docset_dir = Path(docdir.path)
                name = content.get("name") or docset_dir.stem
                title = content.get("title") or name.replace("_", " ")
                # zeal require prefix without any _/' '. this may cause
                # finding in wrong docsets if prefix is the same (ie java_se17,
                # java_se19) but we can't do anything with this
                name = name.partition("_")[0]  # take part before "_"
                keywords = None
                extra = content.get("extra")
                if extra:
                    keywords = extra.get("keywords")

                icon_filename = None
                if (icon := docset_dir.joinpath("icon@2x.png")).is_file():
                    icon_filename = str(icon)

                yield ZealDocset(name, title, keywords, icon_filename)

    def get_icon_name(self):
        return "zeal"

    def provides(self):
        yield ZealDocset

    def should_sort_lexically(self):
        return True


class ZealDocset(Leaf):
    def __init__(self, name, title, keywords, icon):
        super().__init__(name, title)
        self.icon = icon
        self._icon: NonpersistentToken[icons.ComposedIcon] | None = None
        if keywords:
            for alias in keywords:
                self.kupfer_add_alias(alias)

    def get_description(self):
        return _("Zeal %s Docset") % self.name

    def get_gicon(self):
        """Because of we read gicon from file, cache whole icon."""
        if self._icon:
            return self._icon.data

        icon = None
        if self.icon:
            try:
                icon = icons.ComposedIcon(
                    "zeal", icons.get_gicon_from_file(self.icon)
                )
                self._icon = NonpersistentToken(icon)
            except Exception:
                # do not try load icon again
                self.icon = None
            else:
                return icon

        # this is cached in icons
        return icons.ComposedIcon("zeal", "emblem-documents")

    def get_icon_name(self):
        return "zeal"


class ZealSearchInDocset(Action):
    """TextLeaf -> ZealSearchInDocset -> ZealDocset"""

    def __init__(self):
        Action.__init__(self, _("Search In Zeal docset..."))

    def activate(self, leaf, iobj=None, ctx=None):
        assert iobj
        docset = iobj.object
        terms = leaf.object
        launch.spawn_async(["zeal", docset + ":" + terms])

    def item_types(self):
        yield TextLeaf

    def requires_object(self):
        return True

    def object_types(self):
        yield ZealDocset

    def object_source(self, for_item=None):
        return ZealDocsetsSource()

    def get_description(self):
        return _("Search in Zeal docsets")

    def get_gicon(self):
        return icons.ComposedIcon("zeal", "edit-find")

    def get_icon_name(self):
        return "zeal"


class ZealSearchFor(Action):
    """ZealDocset -> ZealSearchFor -> TextLeaf

    This is the opposite action to ZealSearchInDocset
    """

    def __init__(self):
        Action.__init__(self, _("Search For..."))

    def activate(self, leaf, iobj=None, ctx=None):
        assert iobj
        docset = leaf.object
        terms = iobj.object
        launch.spawn_async(["zeal", docset + ":" + terms])

    def item_types(self):
        yield ZealDocset

    def requires_object(self):
        return True

    def object_types(self):
        yield TextLeaf

    def get_description(self):
        return _("Search in Zeal docsets")

    def get_gicon(self):
        return icons.ComposedIcon("zeal", "edit-find")

    def get_icon_name(self):
        return "zeal"



"""
Interfaces that may implements objects.

This file is a part of the program kupfer, which is
released under GNU General Public License v3 (or any later version),
see the main program file, and COPYING for details.
"""

# TODO: use interfaces (?)
# ruff:

import abc

__all__ = ("TextRepresentation", "UriListRepresentation")


# pylint: disable=too-few-public-methods
class TextRepresentation(abc.ABC):  # noqa: B024
    """Kupfer Objects that implement this interface have a plain text
    representation that can be used for Copy & Paste etc."""

    def get_text_representation(self) -> str:
        """The default implementation returns the represented object"""
        # pylint: disable=no-member
        assert hasattr(self, "object")
        return str(self.object)


# pylint: disable=too-few-public-methods
class UriListRepresentation(abc.ABC):  # noqa: B024
    """Kupfer Objects that implement this interface have a uri-list
    representation that can be used for Copy & Paste etc

    get_urilist_representation should return a sequence of string URIs."""

    def get_urilist_representation(self) -> list[str]:
        """The default implementation raises notimplementederror"""
        raise NotImplementedError



__kupfer_name__ = _("Thunar")
__kupfer_sources__ = ("ThunarObjects",)
__kupfer_actions__ = (
    "Reveal",
    "GetInfo",
    "SendTo",
    "CopyTo",
    "LinkTo",
    "MoveTo",
)
__description__ = _("File manager Thunar actions")
__version__ = ""
__author__ = "Ulrik Sverdrup <ulrik.sverdrup@gmail.com>"

import os
import typing as ty
from contextlib import suppress

import dbus

from kupfer import config, plugin_support
from kupfer.obj import (
    Action,
    AppLeaf,
    FileLeaf,
    NotAvailableError,
    RunnableLeaf,
    Source,
)
from kupfer.obj.apps import AppLeafContentMixin
from kupfer.obj.exceptions import InvalidDataError, NoMultiError
from kupfer.support import pretty

if ty.TYPE_CHECKING:
    from gettext import gettext as _

plugin_support.check_dbus_connection()

_SERVICE_NAME = "org.xfce.Thunar"
_OBJECT_PATH = "/org/xfce/FileManager"
_IFACE_NAME = "org.xfce.FileManager"

_TRASH_IFACE_NAME = "org.xfce.Trash"


def _get_thunar():
    """Return the dbus proxy object for Thunar
    we will activate it over d-bus (start if not running)
    """
    bus = dbus.SessionBus()
    try:
        proxy_obj = bus.get_object(_SERVICE_NAME, _OBJECT_PATH)
    except dbus.DBusException as exc:
        pretty.print_error(__name__, exc)
        return None

    return dbus.Interface(proxy_obj, _IFACE_NAME)


def _get_thunar_trash():
    """Return the dbus proxy object for Thunar
    we will activate it over d-bus (start if not running)
    """
    bus = dbus.SessionBus()
    try:
        proxy_obj = bus.get_object(_SERVICE_NAME, _OBJECT_PATH)
    except dbus.DBusException as exc:
        pretty.print_error(__name__, exc)
        return None

    return dbus.Interface(proxy_obj, _TRASH_IFACE_NAME)


def _dummy(*args):
    pass


class Reveal(Action):
    def __init__(self):
        Action.__init__(self, _("Select in File Manager"))

    def wants_context(self):
        return True

    def activate(self, leaf, iobj=None, ctx=None):
        assert ctx

        gfile = leaf.get_gfile()
        parent = gfile.get_parent()
        if not parent:
            return

        uri = parent.get_uri()
        bname = gfile.get_basename()
        id_ = ctx.environment.get_startup_notification_id()
        display = ctx.environment.get_display()
        try:
            # Thunar 1.2 Uses $DISPLAY and $STARTUP_ID args
            _get_thunar().DisplayFolderAndSelect(
                uri,
                bname,
                display,
                id_,
                reply_handler=_dummy,
                error_handler=_dummy,
            )
        except TypeError:
            # Thunar 1.0 Uses $DISPLAY
            _get_thunar().DisplayFolderAndSelect(
                uri, bname, display, reply_handler=_dummy, error_handler=_dummy
            )

    def item_types(self):
        yield FileLeaf


class GetInfo(Action):
    def __init__(self):
        Action.__init__(self, _("Show Properties"))

    def wants_context(self):
        return True

    def activate(self, leaf, iobj=None, ctx=None):
        assert ctx
        gfile = leaf.get_gfile()
        uri = gfile.get_uri()
        id_ = ctx.environment.get_startup_notification_id()
        display = ctx.environment.get_display()
        try:
            # Thunar 1.2 Uses $DISPLAY and $STARTUP_ID args
            _get_thunar().DisplayFileProperties(
                uri, display, id_, reply_handler=_dummy, error_handler=_dummy
            )
        except TypeError:
            # Thunar 1.0 Uses $DISPLAY
            _get_thunar().DisplayFileProperties(
                uri, display, reply_handler=_dummy, error_handler=_dummy
            )

    def item_types(self):
        yield FileLeaf

    def get_description(self):
        return _("Show information about file in file manager")

    def get_icon_name(self):
        return "dialog-information"


class SendTo(Action):
    """Send files to  selected app from "send to" list"""

    def __init__(self):
        Action.__init__(self, _("Send To..."))

    def activate_multiple(self, leaves, iobjs):
        for app in iobjs:
            app.launch(paths=[leaf.object for leaf in leaves])

    def activate(self, leaf, iobj=None, ctx=None):
        assert iobj
        self.activate_multiple((leaf,), (iobj,))

    def item_types(self):
        yield FileLeaf

    def requires_object(self):
        return True

    def object_types(self):
        yield AppLeaf

    def object_source(self, for_item=None):
        return _SendToAppsSource()


def _good_destination(dpath: str, spath: str) -> bool:
    """If directory path @dpath is a valid destination for file @spath
    to be copied or moved to.
    """
    if not os.path.isdir(dpath):
        return False

    spath = os.path.normpath(spath)
    dpath = os.path.normpath(dpath)
    cpfx = os.path.commonprefix((spath, dpath))
    return not os.path.samefile(dpath, spath) and cpfx != spath


def leaf_uri(leaf):
    return leaf.get_gfile().get_uri()


class CopyTo(Action, pretty.OutputMixin):
    def __init__(self):
        Action.__init__(self, _("Copy To..."))

    def wants_context(self):
        return True

    def activate_multiple(self, leaves, iobjects, ctx):
        # Unroll by looping over the destinations,
        # copying everything into each destination
        thunar = _get_thunar()
        work_dir = os.path.expanduser("~/")
        display = ctx.environment.get_display()
        notify_id = ctx.environment.get_startup_notification_id()
        sourcefiles = [leaf_uri(L) for L in leaves]

        def _reply(*args):
            self.output_debug("reply got for copying", *args)

        def _reply_error(exc):
            self.output_debug(exc)
            ctx.register_late_error(NotAvailableError(_("Thunar")))

        for dest_iobj in iobjects:
            desturi = leaf_uri(dest_iobj)
            thunar.CopyInto(
                work_dir,
                sourcefiles,
                desturi,
                display,
                notify_id,
                reply_handler=_reply,
                error_handler=_reply_error,
            )

    def activate(self, leaf, iobj=None, ctx=None):
        return self.activate_multiple([leaf], [iobj], ctx)

    def item_types(self):
        yield FileLeaf

    def valid_for_item(self, leaf):
        return True

    def requires_object(self):
        return True

    def object_types(self):
        yield FileLeaf

    def valid_object(self, obj, for_item):
        return _good_destination(obj.object, for_item.object)

    def get_description(self):
        return _("Copy file to a chosen location")


class MoveTo(Action, pretty.OutputMixin):
    def __init__(self):
        Action.__init__(self, _("Move To..."))

    def wants_context(self):
        return True

    def activate_multiple(self, leaves, iobjects, ctx):
        if len(iobjects) != 1:
            raise NoMultiError

        def _reply():
            self.output_debug("reply got for moving")

        def _reply_error(exc):
            self.output_debug(exc)
            ctx.register_late_error(NotAvailableError(_("Thunar")))

        (dest_iobj,) = iobjects
        # Move everything into the destination
        thunar = _get_thunar()
        if not thunar:
            return

        work_dir = os.path.expanduser("~/")
        display = ctx.environment.get_display()
        notify_id = ctx.environment.get_startup_notification_id()
        sourcefiles = [leaf_uri(L) for L in leaves]
        desturi = leaf_uri(dest_iobj)
        thunar.MoveInto(
            work_dir,
            sourcefiles,
            desturi,
            display,
            notify_id,
            reply_handler=_reply,
            error_handler=_reply_error,
        )

    def activate(self, leaf, iobj=None, ctx=None):
        assert iobj
        assert ctx
        return self.activate_multiple([leaf], [iobj], ctx)

    def item_types(self):
        yield FileLeaf

    def valid_for_item(self, leaf):
        return True

    def requires_object(self):
        return True

    def object_types(self):
        yield FileLeaf

    def valid_object(self, obj, for_item):
        return _good_destination(obj.object, for_item.object)

    def get_description(self):
        return _("Move file to new location")

    def get_icon_name(self):
        return "go-next"


class LinkTo(Action, pretty.OutputMixin):
    def __init__(self):
        Action.__init__(self, _("Symlink In..."))

    def wants_context(self):
        return True

    def activate_multiple(self, leaves, iobjects, ctx):
        # Unroll by looping over the destinations,
        # copying everything into each destination
        thunar = _get_thunar()
        if not thunar:
            return

        work_dir = os.path.expanduser("~/")
        display = ctx.environment.get_display()
        notify_id = ctx.environment.get_startup_notification_id()
        sourcefiles = [leaf_uri(L) for L in leaves]

        def _reply(*args):
            self.output_debug("reply got for copying", *args)

        def _reply_error(exc):
            self.output_debug(exc)
            ctx.register_late_error(NotAvailableError(_("Thunar")))

        for dest_iobj in iobjects:
            desturi = leaf_uri(dest_iobj)
            thunar.LinkInto(
                work_dir,
                sourcefiles,
                desturi,
                display,
                notify_id,
                reply_handler=_reply,
                error_handler=_reply_error,
            )

    def activate(self, leaf, iobj=None, ctx=None):
        assert iobj
        assert ctx
        return self.activate_multiple([leaf], [iobj], ctx)

    def item_types(self):
        yield FileLeaf

    def valid_for_item(self, leaf):
        return True

    def requires_object(self):
        return True

    def object_types(self):
        yield FileLeaf

    def valid_object(self, obj, for_item):
        return _good_destination(obj.object, for_item.object)

    def get_description(self):
        return _("Create a symlink to file in a chosen location")


class OpenTrash(RunnableLeaf):
    def __init__(self):
        RunnableLeaf.__init__(self, None, _("Open Trash"))

    def wants_context(self):
        return True

    def run(self, ctx=None):
        assert ctx
        id_ = ctx.environment.get_startup_notification_id()
        thunar = _get_thunar_trash()
        try:
            # Thunar 1.2 Uses $DISPLAY and $STARTUP_ID args
            thunar.DisplayTrash(
                ctx.environment.get_display(),
                id_,
            )
        except TypeError:
            # Thunar 1.0 uses only $DISPLAY arg
            thunar.DisplayTrash(
                ctx.environment.get_display(),
            )

    def get_description(self):
        return _("Open Trash in Thunar")

    def get_icon_name(self):
        return "user-trash"


class EmptyTrash(RunnableLeaf):
    def __init__(self):
        RunnableLeaf.__init__(self, None, _("Empty Trash"))

    def wants_context(self):
        return True

    def run(self, ctx=None):
        assert ctx
        id_ = ctx.environment.get_startup_notification_id()
        thunar = _get_thunar_trash()
        try:
            # Thunar 1.2 Uses $DISPLAY and $STARTUP_ID args
            thunar.EmptyTrash(
                ctx.environment.get_display(),
                id_,
                reply_handler=_dummy,
                error_handler=_dummy,
            )
        except TypeError:
            # Thunar 1.0 uses only $DISPLAY arg
            thunar.EmptyTrash(
                ctx.environment.get_display(),
                reply_handler=_dummy,
                error_handler=_dummy,
            )

    def get_description(self):
        return None

    def get_icon_name(self):
        return "user-trash-full"


class ThunarObjects(AppLeafContentMixin, Source):
    appleaf_content_id = ("Thunar", "thunar")
    source_scan_interval: int = 36000
    serializable = None

    def __init__(self):
        Source.__init__(self, _("Thunar"))

    def get_items(self):
        yield EmptyTrash()
        yield OpenTrash()

    def provides(self):
        yield RunnableLeaf

    def get_icon_name(self):
        return "Thunar"


class _SendToAppsSource(Source):
    """Send To items source"""

    def __init__(self):
        Source.__init__(self, _("Thunar Send To Objects"))

    def get_items(self):
        for data_dir in config.get_data_dirs("sendto", package="Thunar"):
            with os.scandir(data_dir) as sd:
                for dir_entry in sd:
                    if not dir_entry.is_file() or not dir_entry.name.endswith(
                        ".desktop"
                    ):
                        continue

                    with suppress(InvalidDataError):
                        yield AppLeaf(
                            init_path=dir_entry.path, require_x=False
                        )

    def get_icon_name(self):
        return "Thunar"

    def provides(self):
        yield AppLeaf



import atexit
import os
import sys
import typing as ty

from gi.repository import Gtk

from kupfer.obj import RunnableLeaf, Source
from kupfer.obj.apps import AppLeafContentMixin
from kupfer.support import pretty
from kupfer.ui import about, kupferhelp, preferences
from kupfer.version import DESKTOP_ID

if ty.TYPE_CHECKING:
    from gettext import gettext as _


__kupfer_sources__ = ("KupferSource",)
__kupfer_actions__ = ()
__author__ = "Ulrik Sverdrup <ulrik.sverdrup@gmail.com>"

__all__ = list(__kupfer_sources__)


def _is_debug():
    # Return True if Kupfer is in debug mode
    return pretty.DEBUG


class DebugRestart(RunnableLeaf):
    def __init__(self):
        RunnableLeaf.__init__(self, None, _("Restart Kupfer"))

    @classmethod
    def _exec_new_kupfer(cls, executable, argv):
        os.execvp(executable, [executable, *argv])

    def run(self, ctx=None):
        Gtk.main_quit()
        atexit.register(self._exec_new_kupfer, sys.executable, sys.argv)

    def get_description(self):
        return str(self)

    def get_icon_name(self):
        return "view-refresh"


class Quit(RunnableLeaf):
    qf_id = "quit"

    def __init__(self, name=None):
        super().__init__(name=name or _("Quit"))

    def run(self, ctx=None):
        Gtk.main_quit()

    def get_description(self):
        return _("Quit Kupfer")

    def get_icon_name(self):
        return "application-exit"


class About(RunnableLeaf):
    def __init__(self, name=None):
        if not name:
            name = _("About Kupfer")
        super().__init__(name=name)

    def wants_context(self):
        return True

    def run(self, ctx=None):
        assert ctx
        about.show_about_dialog(ctx.environment)

    def get_description(self):
        return _("Show information about Kupfer authors and license")

    def get_icon_name(self):
        return "help-about"


class Help(RunnableLeaf):
    def __init__(self, name=None):
        super().__init__(name=name or _("Kupfer Help"))

    def wants_context(self):
        return True

    def run(self, ctx=None):
        assert ctx
        kupferhelp.show_help(ctx.environment)

    def get_description(self):
        return _("Get help with Kupfer")

    def get_icon_name(self):
        return "help-contents"


class Preferences(RunnableLeaf):
    def __init__(self, name=None):
        super().__init__(name=name or _("Kupfer Preferences"))

    def wants_context(self):
        return True

    def run(self, ctx=None):
        assert ctx
        preferences.show_preferences(ctx.environment)

    def get_description(self):
        return _("Show preferences window for Kupfer")

    def get_icon_name(self):
        return "preferences-desktop"


class KupferSource(AppLeafContentMixin, Source):
    appleaf_content_id = DESKTOP_ID

    def __init__(self, name=None):
        Source.__init__(self, name or _("Kupfer"))

    def is_dynamic(self):
        return True

    def get_items(self):
        yield Preferences()
        yield Help()
        yield About()
        yield Quit()
        yield DebugRestart()

    def get_description(self):
        return None

    def get_icon_name(self):
        return "search"

    def provides(self):
        yield RunnableLeaf



#!/usr/bin/python3
"""
kupfer      A convenient command and access tool

Copyright 2007-–2023 Ulrik Sverdrup and other Kupfer authors

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

if __name__ == "__main__":
    from pathlib import Path

    from kupfer import main

    # if kupfer is installed in this same dir should be defaults.cfg
    if not Path(__file__).with_name("defaults.cfg").is_file():
        print("!!!!!!!!!!!!!! WARN !!!!!!!!!!!!!!")
        print(
            "Launching Kupfer by kupfer.py is dedicated only for development. "
            "This may totally broke some plugins, and some parts may not not "
            "work as expected."
        )
        print("Also Kupfer may run much slower.")
        print("Please install and run Kupfer as described in documentation.")
        print("!!!!!!!!!!!!!! WARN !!!!!!!!!!!!!!")

    main.main()



#! /usr/bin/env python3

"""
About dialog.
"""

from __future__ import annotations

import typing as ty

from gi.repository import Gtk

from kupfer import version

if ty.TYPE_CHECKING:
    from kupfer.ui.uievents import GUIEnvironmentContext


# pylint: disable=too-few-public-methods
class _AboutDialog:
    _instance: _AboutDialog | None = None

    @classmethod
    def instance(cls) -> _AboutDialog:
        if cls._instance is None:
            cls._instance = _AboutDialog()

        assert cls._instance
        return cls._instance

    def __init__(self):
        self._dlg = abdlg = Gtk.AboutDialog()
        abdlg.set_program_name(version.PROGRAM_NAME)
        abdlg.set_icon_name(version.ICON_NAME)
        abdlg.set_logo_icon_name(version.ICON_NAME)
        abdlg.set_version(version.VERSION)
        abdlg.set_comments(version.SHORT_DESCRIPTION)
        abdlg.set_copyright(version.COPYRIGHT)
        abdlg.set_website(version.WEBSITE)
        abdlg.set_license(version.LICENSE)
        abdlg.set_authors(version.AUTHORS)
        if version.DOCUMENTERS:
            abdlg.set_documenters(version.DOCUMENTERS)

        if version.TRANSLATOR_CREDITS:
            abdlg.set_translator_credits(version.TRANSLATOR_CREDITS)

        if version.ARTISTS:
            abdlg.set_artists(version.ARTISTS)

        abdlg.connect("response", self._on_response)

    def show(self, ctxenv: GUIEnvironmentContext | None = None) -> None:
        if ctxenv:
            ctxenv.present_window(self._dlg)
        else:
            self._dlg.present()

    def _on_response(self, dialog: Gtk.Dialog, response_id: ty.Any) -> bool:
        dialog.destroy()
        _AboutDialog._instance = None
        return True


def show_about_dialog(
    ctxenv: GUIEnvironmentContext | None = None,
) -> None:
    """Create an about dialog and show it."""
    dlg = _AboutDialog.instance()
    dlg.show()



# TRANS: "Glob" is the matching files like a shell with "*.py" etc.
__kupfer_name__ = _("Glob")
__kupfer_actions__ = ("Glob",)
__description__ = _("Select objects using '*' and '?' as wildcards.")
__version__ = ""
__author__ = ""

import fnmatch
import re
import typing as ty

from kupfer.obj import Action, Leaf, OperationError, TextLeaf, TextSource
from kupfer.obj.compose import MultipleLeaf

if ty.TYPE_CHECKING:
    from gettext import gettext as _


class Glob(Action):
    def __init__(self):
        Action.__init__(self, _("Glob"))

    def activate(self, leaf, iobj=None, ctx=None):
        assert iobj
        return self.activate_multiple((leaf,), (iobj,))

    def activate_multiple(self, objects, iobjects):
        ## Do case-insensitive matching
        ## As a special case, understand '**/' prefix as recursive

        def get_subcatalog_matches(subcatalog, pat, recursive, paths):
            if len(paths) > 1000:  # noqa:PLR2004
                raise OperationError("Globbing way too many objects")

            for content in subcatalog.content_source().get_leaves():
                if recursive and content.has_content():
                    get_subcatalog_matches(content, pat, recursive, paths)
                elif re.match(pat, str(content), flags=re.IGNORECASE):
                    paths.append(content)

        paths: list[str] = []
        for iobj in iobjects:
            glob = iobj.object
            if glob.startswith("**/"):
                glob = glob[3:]
                recursive = True
            else:
                recursive = False

            pat = fnmatch.translate(glob)
            for obj in objects:
                get_subcatalog_matches(obj, pat, recursive, paths)

        if paths:
            return MultipleLeaf(paths)

        return None

    def has_result(self):
        return True

    def item_types(self):
        yield Leaf

    def valid_for_item(self, leaf):
        return leaf.has_content()

    def requires_object(self):
        return True

    def object_types(self):
        yield TextLeaf

    def object_source(self, for_item=None):
        return TextSource()

    def valid_object(self, iobj, for_item):
        return ("*" in iobj.object) or ("?" in iobj.object)

    def get_description(self):
        return __description__



# note: not used

from __future__ import annotations

import functools
import typing as ty

# NOTE: was glib; changed to gio.repo.GLib
from gi.repository import GLib, Gtk

from kupfer import config, version

T = ty.TypeVar("T")


def idle_call(func: ty.Callable[..., T]) -> ty.Callable[..., T]:
    @functools.wraps(func)
    def idle_wrapper(*args):
        GLib.idle_add(func, *args)  # pylint: disable=no-member

    return idle_wrapper


_HEADER_MARKUP = '<span weight="bold" size="larger">%s</span>'


class ProgressDialogController:
    def __init__(
        self, title: str, header: str | None = None, max_value: int = 100
    ):
        """Create a new progress dialog

        @header: first line of dialog

        The methods show, hide and update are all wrapped to be
        safe to call from any thread.
        """
        self.aborted = False
        self._max_value = float(max_value)
        ui_file = config.get_data_file("progress_dialog.ui")
        self._construct_dialog(ui_file, title, header)

    @idle_call
    def _construct_dialog(self, ui_file: str, title: str, header: str) -> None:
        builder = Gtk.Builder()
        builder.set_translation_domain(version.PACKAGE_NAME)

        builder.add_from_file(ui_file)
        builder.connect_signals(self)  # pylint: disable=no-member
        self._window = builder.get_object("window_progress")
        self._button_abort = builder.get_object("button_abort")
        self._progressbar = builder.get_object("progressbar")
        self._label_info = builder.get_object("label_info")

        label_header = builder.get_object("label_header")
        if header:
            label_header.set_markup(
                _HEADER_MARKUP % GLib.markup_escape_text(header)
            )
        else:
            label_header.hide()

        self._window.set_title(title)
        self.update(0, "", "")

    def on_button_abort_clicked(self, widget: Gtk.Button) -> None:
        self.aborted = True
        self._button_abort.set_sensitive(False)

    @idle_call
    def show(self) -> None:
        self._window.present()

    @idle_call
    def hide(self) -> None:
        self._window.hide()

    @idle_call
    def update(self, value: float, label: str, text: str) -> bool:
        """Update dialog information.

        @value: value to set for progress bar
        @label: current action (displayed in emphasized style)
        @text: current information (normal style)

        @return true when abort button was pressed
        """
        self._progressbar.set_fraction(min(value / self._max_value, 1.0))
        self._label_info.set_markup(
            "<b>"
            + GLib.markup_escape_text(label)
            + "</b>"
            + GLib.markup_escape_text(text)
        )
        return self.aborted



#!/usr/bin/python3
"""
kupfer      A convenient command and access tool.

This is launcher that load some useful libraries for development.

Copyright 2007-2023 Ulrik Sverdrup and other Kupfer authors

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import sys

try:
    import stackprinter

    stackprinter.set_excepthook(
        style="color",
        suppressed_paths=[r"*/site-packages/typeguard/"],
    )
except ImportError:
    try:
        from rich.traceback import install

        suppress_modules = []

        try:
            import typeguard

            suppress_modules.append(typeguard)
        except ImportError:
            pass

        install(show_locals=True, suppress=suppress_modules)
        print("rich.traceback installed")
    except ImportError:
        pass

try:
    import icecream

    icecream.install()
    icecream.ic.configureOutput(includeContext=True)

    import traceback

    def ic_stack(*args, **kwargs):
        stack = "".join(tbs.rstrip() for tbs in traceback.format_stack()[:-2])
        ic(stack, *args, **kwargs)  # type: ignore

    import inspect

    class ShiftedIceCreamDebugger(icecream.IceCreamDebugger):  # type: ignore
        def format(self, *args):
            # one more frame back
            call_frame = inspect.currentframe().f_back.f_back  # type: ignore
            return self._format(call_frame, *args)

    sic = ShiftedIceCreamDebugger()

    def ic_trace(func):
        def wrapper(*args, **kwargs):
            sic(func, args, kwargs)
            res = func(*args, **kwargs)
            sic(func, res)
            return res

        return wrapper

    import builtins

    builtins.ic_stack = ic_stack  # type: ignore
    builtins.ic_trace = ic_trace  # type: ignore
except ImportError:  # Graceful fallback if IceCream isn't installed.
    pass


try:
    if "--debug" in sys.argv:
        from typeguard import install_import_hook

        install_import_hook("kupfer")
        print("WARN! typeguard hook installed")

        import typing

        typing.TYPE_CHECKING = True  # type: ignore

        import typeguard._checkers as checkers

        checkers.check_protocol = None  # agronholm/typeguard#465
except ImportError as err:
    print(err)

try:
    from pympler import tracker
except ImportError:
    tracker = None


if __name__ == "__main__":
    from kupfer import main

    if tracker:
        tr = tracker.SummaryTracker()

    print("!!!!!!!!!!!!!! WARN !!!!!!!!!!!!!!")
    print(
        "Launching Kupfer by kupfer_dev.py is dedicated only for development. "
        "This may totally broke some plugins, and some parts may not not "
        "work as expected."
    )
    print("Also Kupfer may run much slower.")
    print("Please install and run Kupfer as described in documentation.")
    print("!!!!!!!!!!!!!! WARN !!!!!!!!!!!!!!")

    main.main()

    if tracker:
        print("--- TRACKER start ---")
        tr.print_diff()
        print("--- TRACKER end ---")



import json
import os
import typing as ty

_CONTAINER = "folder"
_UNWANTED_SCHEME = ("data", "place", "javascript")


def _is_container(item: dict[str, ty.Any]) -> bool:
    return item["type"] == _CONTAINER  # type:ignore


def _is_good_bookmark(item: dict[str, ty.Any]) -> bool:
    if url := item.get("url"):
        return url.split(":", 1)[0] not in _UNWANTED_SCHEME

    return False


def get_bookmarks(bookmarks_file: str) -> ty.Iterable[dict[str, ty.Any]]:
    # construct and configure the parser
    if not bookmarks_file:
        return

    with open(bookmarks_file, encoding="UTF-8") as fin:
        content = fin.read()
        root = json.loads(content)

    folders = []

    # add some folders
    folders.extend(root["roots"]["bookmark_bar"]["children"])
    folders.extend(root["roots"]["other"]["children"])

    # make a dictionary of unique bookmarks
    bmap = set()
    for item in folders:
        if _is_good_bookmark(item) and (iid := item["id"]) not in bmap:
            bmap.add(iid)
            yield item

        if _is_container(item):
            folders.extend(item["children"])


if __name__ == "__main__":
    fpath = os.path.expanduser("~/.config/chromium/Default/")
    print("Parsed # bookmarks:", len(list(get_bookmarks(fpath))))



from __future__ import annotations

import os
import pickle
import random
import time
import typing as ty
from collections import defaultdict
from pathlib import Path

from kupfer import config
from kupfer.support import conspickle, pretty

if ty.TYPE_CHECKING:
    from kupfer.obj.base import Action, KupferObject, Leaf

__all__ = (
    "add_favorite",
    "erase_object_affinity",
    "get_correlation_bonus",
    "get_object_has_affinity",
    "get_record_score",
    "is_favorite",
    "load",
    "record_action_activations",
    "record_search_hit",
    "remove_favorite",
    "replace_favorites",
    "save",
    "set_correlation",
    "unregister",
)

_MNEMONICS_FILENAME = "mnemonics.pickle"
_CORRELATION_KEY: ty.Final = "kupfer.bonus.correlation"
_ACTIVATIONS_KEY: ty.Final = "kupfer.bonus.activations"

## this is a harmless default
_DEFAULT_ACTIONS: ty.Final = {
    "<kupfer.obj.apps.AppLeaf gnome-terminal>": "<kupfer.obj.apps.LaunchAgain>",
    "<kupfer.obj.apps.AppLeaf xfce4-terminal>": "<kupfer.obj.apps.LaunchAgain>",
}

## Favorites is set of favorites (repr(obj))
_FAVORITES: ty.Final[set[str]] = set()
## _PLUG_FAVS are favorites by plugin; to use must be merged to _FAVORITES
_PLUG_FAVS: ty.Final[dict[str, list[str]]] = {}


class Mnemonics:
    """Class to describe a collection of mnemonics as well as the total count."""

    __slots__ = ("count", "last_ts_used", "mnemonics")

    def __init__(self) -> None:
        self.mnemonics: defaultdict[str, int] = defaultdict(int)
        self.count: int = 0
        self.last_ts_used: int = 0

    def __repr__(self) -> str:
        mnm = ", ".join(f"{m}:{c}" for m, c in self.mnemonics.items())
        return (
            f"<{self.__class__.__name__} cnt={self.count} mnm={mnm}"
            f" ts={self.last_ts_used}>"
        )

    def increment(self, mnemonic: str | None = None) -> None:
        if mnemonic:
            self.mnemonics[mnemonic] += 1

        self.count += 1
        self.last_ts_used = int(time.time())

    def decrement(self) -> None:
        """Decrement total count and the least mnemonic"""
        if self.mnemonics:
            key = min(self.mnemonics.keys(), key=lambda k: self.mnemonics[k])
            if (mcount := self.mnemonics[key]) > 1:
                self.mnemonics[key] = mcount - 1
            else:
                self.mnemonics.pop(key)

        self.count = max(self.count - 1, 0)

    def __bool__(self) -> bool:
        return self.count > 0

    def __getstate__(self) -> dict[str, ty.Any]:
        return {
            "count": self.count,
            "mnemonics": dict(self.mnemonics),
            "last_ts_used": self.last_ts_used,
        }

    def __setstate__(self, state: dict[str, ty.Any]) -> None:
        self.count = state.get("count", 0)
        self.last_ts_used = state.get("last_ts_used", 0)
        self.mnemonics = defaultdict(int)
        self.mnemonics.update(state.get("mnemonics", {}))


class Learning:
    @classmethod
    def unpickle_register(cls, pickle_file: str) -> dict[str, ty.Any] | None:
        try:
            pfile = Path(pickle_file).read_bytes()
            data = conspickle.ConservativeUnpickler.loads(pfile)
            assert isinstance(data, dict), "Stored object not a dict"
            pretty.print_debug(__name__, f"Reading from {pickle_file}")
            return data
        except OSError:
            pass
        except (pickle.PickleError, Exception) as exc:
            pretty.print_error(__name__, f"Error loading {pickle_file}: {exc}")

        return None

    @classmethod
    def pickle_register(cls, reg: dict[str, ty.Any], pickle_file: str) -> bool:
        ## Write to tmp then rename over for atomicity
        tmp_pickle_file = f"{pickle_file}.{os.getpid()}"
        pretty.print_debug(__name__, f"Saving to {pickle_file}")
        Path(tmp_pickle_file).write_bytes(
            pickle.dumps(reg, pickle.HIGHEST_PROTOCOL)
        )
        os.rename(tmp_pickle_file, pickle_file)
        return True


# under _CORRELATION_KEY is {str:str}:
#   map repr(action) -> repr(object)
# under _ACTIVATIONS_KEY is {str:(str, int)}:
#   map repr(action) -> (repr(leaf),last usage timestamp)
# other keys keeps Mnemonics
_REGISTER: dict[
    str, Mnemonics | dict[str, str] | dict[str, tuple[str, int]]
] = {}


def _get_register_mnemonics() -> ty.Iterator[tuple[str, Mnemonics]]:
    for leaf, mne in _REGISTER.items():
        if isinstance(mne, Mnemonics):
            yield leaf, mne


def record_search_hit(obj: ty.Any, key: str | None = None) -> None:
    """Record that KupferObject @obj was used, with the optional
    search term @key recording.
    When key is None - skip registration (this is only valid when action is
    performed by accelerator)."""
    if key is None:
        return

    name = repr(obj)
    mns = _REGISTER.get(name)
    if not mns:
        mns = _REGISTER[name] = Mnemonics()

    assert isinstance(mns, Mnemonics)
    mns.increment(key or "")


def get_record_score(obj: ty.Any, key: str = "") -> int:
    """Get total score for KupferObject @obj, bonus score is given for @key
    matches"""
    name = repr(obj)
    fav = 7 if name in _FAVORITES else 0

    mns = _REGISTER.get(name)
    if mns is None:
        return fav

    assert isinstance(mns, Mnemonics)
    if not key:
        return fav + 50 - int(50.0 / (mns.count + 1))

    stats = mns.mnemonics
    closescr = sum(val for m, val in stats.items() if m.startswith(key))
    exact = stats[key]
    mnscore = 80 - int(50.0 / (closescr + 1) + 30.0 / (exact + 1))
    return fav + mnscore


def get_correlation_bonus(obj: Action, for_leaf: Leaf | None) -> int:
    """Get the bonus rank for @obj when used with @for_leaf."""

    # favorites
    rval = ty.cast("dict[str, str]", _REGISTER[_CORRELATION_KEY])
    repr_obj = repr(obj)
    repr_leaf = repr(for_leaf)
    if rval.get(repr_leaf) == repr_obj:
        return 50

    raval = ty.cast("dict[str, tuple[str, int]]", _REGISTER[_ACTIVATIONS_KEY])

    # bonus for last used action for object
    if (val := raval.get(repr_leaf)) and val[0] == repr_obj:
        return 20

    # bonus for last used action for object type
    if (val := raval.get(repr(type(for_leaf)))) and val[0] == repr_obj:
        return 7

    return 0


def set_correlation(obj: Action, for_leaf: Leaf) -> None:
    """Register @obj to get a bonus when used with @for_leaf."""
    rval = ty.cast("dict[str, str]", _REGISTER[_CORRELATION_KEY])
    rval[repr(for_leaf)] = repr(obj)


def record_action_activations(obj: Action, for_leaf: Leaf) -> None:
    """Record action activation for leaf that boost this action in next search.
    Also registered is object class so using this action for similar object
    also get some (smaller) bonus."""

    rval = ty.cast("dict[str, tuple[str, int]]", _REGISTER[_ACTIVATIONS_KEY])

    repr_obj = repr(obj)
    now = int(time.time())
    rval[repr(for_leaf)] = (repr_obj, now)
    rval[repr(type(for_leaf))] = (repr_obj, now)


def get_object_has_affinity(obj: Leaf) -> bool:
    """Return if @obj has any positive score in the register."""
    robj = repr(obj)
    return bool(
        _REGISTER.get(robj) or _REGISTER[_CORRELATION_KEY].get(robj)  # type: ignore
    )


def erase_object_affinity(obj: Leaf) -> None:
    """Remove all track of affinity for @obj."""
    robj = repr(obj)
    _REGISTER.pop(robj, None)
    _REGISTER[_CORRELATION_KEY].pop(robj, None)  # type: ignore


def _prune_register(goalitems: int = 500) -> None:
    """Try to reduce number of mnemonic to `goalitems`.

    in first pass try to delete oldest mnemonics with score == 1.

    Then, remove items with chance (len/25000)

    Assuming homogeneous records (all with score one) we keep:
    x_n+1 := x_n * (1 - chance)

    To this we have to add the expected number of added mnemonics per
    invocation, est. 10, and we can estimate a target number of saved mnemonics.
    """

    # get all items sorted by last used time
    items: list[tuple[int, str, Mnemonics]] = sorted(
        (mne.last_ts_used, leaf, mne)
        for leaf, mne in _get_register_mnemonics()
    )
    to_del = []
    to_del_cnt = len(items) - goalitems
    for _ts, leaf, mne in items:
        mne.decrement()
        if not mne:
            to_del.append(leaf)
            to_del_cnt -= 1
            if not to_del_cnt:
                break

    for leaf in to_del:
        _REGISTER.pop(leaf)

    pretty.print_debug(
        __name__,
        f"Pruned register ({len(_REGISTER)} mnemonics, {len(to_del)} oldest deleted)",
    )

    if len(_REGISTER) <= goalitems:
        return

    random.seed()
    rand = random.random

    flux = 2.0
    alpha = flux / goalitems**2

    chance = min(0.1, len(_REGISTER) * alpha)
    to_del = []
    for leaf, mne in _get_register_mnemonics():
        if rand() <= chance:
            assert isinstance(mne, Mnemonics)
            mne.decrement()
            if not mne:
                to_del.append(leaf)

    for leaf in to_del:
        _REGISTER.pop(leaf)

    pretty.print_debug(
        __name__,
        f"Pruned register ({len(_REGISTER)} mnemonics, {len(to_del)} deleted)",
    )


def _purge_action_reg(goalitems: int) -> None:
    """Purge action usage - remove oldest items up to `goalitems` count"""
    raval = ty.cast("dict[str, tuple[str, int]]", _REGISTER[_ACTIVATIONS_KEY])
    if len(raval) <= goalitems:
        return

    to_delkv = sorted((ts, key) for key, (_obj, ts) in raval.items())
    # delete oldest entries up to goalitems
    for _ts, key in to_delkv[: len(raval) - goalitems]:
        raval.pop(key)


def load() -> None:
    """Load learning database."""
    _REGISTER.clear()

    if (filepath := config.get_config_file(_MNEMONICS_FILENAME)) and (
        reg := Learning.unpickle_register(filepath)
    ):
        _REGISTER.update(reg)

    if _CORRELATION_KEY not in _REGISTER:
        _REGISTER[_CORRELATION_KEY] = _DEFAULT_ACTIONS

    if _ACTIVATIONS_KEY not in _REGISTER:
        _REGISTER[_ACTIVATIONS_KEY] = ty.cast("dict[str, tuple[str, int]]", {})


_MAX_REGISTER_SIZE: ty.Final[int] = 500


def save() -> None:
    """Save the learning record."""
    if not _REGISTER:
        pretty.print_debug(__name__, "Not writing empty register")
        return

    _purge_action_reg(_MAX_REGISTER_SIZE)
    if len(_REGISTER) > _MAX_REGISTER_SIZE:
        _prune_register(_MAX_REGISTER_SIZE)

    filepath = config.save_config_file(_MNEMONICS_FILENAME)
    assert filepath
    Learning.pickle_register(_REGISTER, filepath)


def _rebuild_favorites():
    """Build _FAVORITES set from _PLUG_FAVS buckets."""
    _FAVORITES.clear()
    for favs in _PLUG_FAVS.values():
        _FAVORITES.update(favs)


def add_favorite(plugin_id: str, *objs: KupferObject) -> None:
    """Add favorites `objs` to `plugin_id` bucket."""
    fset = _PLUG_FAVS.get(plugin_id)
    if fset:
        # filter and add new obj
        nfset = [obj for obj in map(repr, objs) if obj not in fset]
        fset.extend(nfset)
    else:
        nfset = _PLUG_FAVS[plugin_id] = list(map(repr, objs))

    _FAVORITES.update(nfset)


def replace_favorites(plugin_id: str, *objs: KupferObject) -> None:
    """Replace favorites in `plugin_id` bucket. If no `objs` remove all
    favorites for `plugin_id`."""
    if not objs:
        if plugin_id in _PLUG_FAVS:
            del _PLUG_FAVS[plugin_id]
            _rebuild_favorites()

        return

    fset = _PLUG_FAVS[plugin_id] = list(map(repr, objs))
    _FAVORITES.update(fset)


def remove_favorite(plugin_id: str, obj: KupferObject) -> None:
    """Remove `obj` from favorites in `plugin_id` bucket."""
    if fset := _PLUG_FAVS.get(plugin_id):
        try:
            fset.remove(repr(obj))
        except ValueError:
            return

        if fset:
            _PLUG_FAVS[plugin_id] = fset
        else:
            del _PLUG_FAVS[plugin_id]

        _rebuild_favorites()


def is_favorite(obj: KupferObject) -> bool:
    return repr(obj) in _FAVORITES


def unregister(obj):
    _REGISTER.pop(obj, None)



__kupfer_name__ = _("Show Text")
__kupfer_actions__ = ("ShowText", "LargeType", "ShowNotification")
__description__ = _("Display text in a window")
__version__ = ""
__author__ = "US"
import typing as ty

from kupfer import icons
from kupfer.obj import Action, Leaf, TextLeaf
from kupfer.support import textutils
from kupfer.ui import uiutils

if ty.TYPE_CHECKING:
    from gettext import gettext as _


class ShowText(Action):
    rank_adjust = -10

    def __init__(self):
        Action.__init__(self, _("Show Text"))

    def wants_context(self):
        return True

    def activate(self, leaf, iobj=None, ctx=None):
        assert ctx
        uiutils.show_text_result(
            leaf.get_text_representation(), title=_("Show Text"), ctx=ctx
        )

    def item_types(self):
        yield TextLeaf

    def get_description(self):
        return _("Display text in a window")

    def get_icon_name(self):
        return "format-text-bold"


class LargeType(Action):
    rank_adjust = -10

    def __init__(self):
        Action.__init__(self, _("Large Type"))

    def wants_context(self):
        return True

    def activate(self, leaf, iobj=None, ctx=None):
        assert ctx
        return self.activate_multiple((leaf,), ctx)

    def activate_multiple(self, objects, ctx):
        all_texts = [obj.get_text_representation() for obj in objects]
        uiutils.show_large_type("\n".join(all_texts), ctx)

    def item_types(self):
        yield Leaf

    def valid_for_item(self, leaf):
        return hasattr(leaf, "get_text_representation")

    def get_description(self):
        return _("Display text in a window")

    def get_gicon(self):
        return icons.ComposedIcon.new("format-text-bold", "zoom-in")

    def get_icon_name(self):
        return "format-text-bold"


class ShowNotification(Action):
    rank_adjust = -10

    def __init__(self):
        Action.__init__(self, _("Show Notification"))

    def activate(self, leaf, iobj=None, ctx=None):
        title, body = textutils.extract_title_body(leaf.object)
        if body:
            uiutils.show_notification(
                title, body, icon_name=self.get_icon_name()
            )
        else:
            uiutils.show_notification(title)

    def item_types(self):
        yield TextLeaf

    def get_icon_name(self):
        return "format-text-bold"



"""
This plugin for simplicity use cli interface instead of api, so we don't need
additional modules.
"""

from __future__ import annotations

__kupfer_name__ = _("Tmux / Tmuxp")
__kupfer_sources__ = ("TmuxSessionsSource", "TmuxpWorkspacesSource")
__description__ = _("Manage tmux and tmuxp sessions")
__version__ = "2023.1"
__author__ = "Karol Będkowski"

import datetime
import os
import typing as ty

from kupfer import launch, plugin_support
from kupfer.obj import Action, Leaf, Source
from kupfer.obj.helplib import FileMonitorToken, FilesystemWatchMixin

if ty.TYPE_CHECKING:
    from gettext import gettext as _

try:
    import libtmux
except ImportError:
    libtmux = None

plugin_support.check_command_available("tmux")


class TmuxSession(Leaf):
    """Represented object is the session_id as string"""

    def __init__(
        self, sid: str, name: str, attached: str, created: str
    ) -> None:
        super().__init__(sid, name)
        self._attached = attached != "0"
        self._created = datetime.datetime.fromtimestamp(int(created))

    def get_actions(self):
        yield Attach()

    def get_description(self) -> str:
        return _("%(status)s tmux session, created %(time)s") % {
            "status": _("Attached") if self._attached else _("Detached"),
            "time": self._created,
        }

    def get_icon_name(self):
        return "gnome-window-manager"


def tmux_sessions(
    session_id: str | None = None,
) -> ty.Iterator[ty.Collection[str]]:
    if libtmux and (srv := libtmux.Server()):
        for sess in srv.sessions:
            if not sess.session_id:
                continue

            yield (
                sess.session_id,
                sess.session_name or f"session {sess.session_id}",
                sess.session_attached or "",
                sess.session_created or "",
            )

        return

    # fallback
    cmd = (
        "tmux list-sessions -F "
        "'#{session_id}\t#{session_name}\t"
        "#{session_attached}\t#{session_created}'"
    )
    if session_id is not None:
        cmd += f" -f '#{{m:{session_id},#{{session_id}}}}'"

    with os.popen(cmd) as pipe:
        output = pipe.read()

    for line in output.splitlines():
        yield line.split("\t")


class TmuxSessionsSource(Source):
    """Source for tmux sessions"""

    source_use_cache = False

    def __init__(self):
        super().__init__(_("tmux Sessions"))

    def is_dynamic(self):
        return True

    def get_items(self):
        for session in tmux_sessions():
            yield TmuxSession(*session)

    def get_description(self):
        return _("Active tmux sessions")

    def get_icon_name(self):
        return "terminal"

    def provides(self):
        yield TmuxSession


class Attach(Action):
    def __init__(self):
        name = _("Attach")
        super().__init__(name)

    def activate(self, leaf, iobj=None, ctx=None):
        sid = leaf.object
        action_argv = ["tmux", "attach-session", "-t", sid, "-d"]
        launch.spawn_in_terminal(action_argv)


class TmuxpSession(Leaf):
    def __init__(self, name: str) -> None:
        super().__init__(name, name.capitalize())

    def get_actions(self):
        yield StartTmuxpSession()

    def get_description(self) -> str:
        return _("tmuxp saved session")

    def get_icon_name(self):
        return "gnome-window-manager"


class StartTmuxpSession(Action):
    def __init__(self):
        name = _("Start session")
        super().__init__(name)

    def activate(self, leaf, iobj=None, ctx=None):
        action_argv = ["tmuxp", "load", leaf.object]
        launch.spawn_in_terminal(action_argv)


class TmuxpWorkspacesSource(Source, FilesystemWatchMixin):
    """Source for tmuxp workspaces"""

    source_scan_interval: int = 3600
    _tmuxp_home = "~/.tmuxp/"

    def __init__(self):
        super().__init__(_("tmuxp Workspaces"))
        self._monitor_token: FileMonitorToken | None = None

    def initialize(self):
        tmux_ws_dir = os.path.expanduser(self._tmuxp_home)
        self._monitor_token = self.monitor_directories(tmux_ws_dir, force=True)

    def finalize(self):
        self.stop_monitor_fs_changes(self._monitor_token)

    def get_items(self):
        with os.popen("tmuxp ls") as pipe:
            output = pipe.read()

        for line in output.splitlines():
            if line := line.strip():
                yield TmuxpSession(line)

    def get_description(self):
        return _("Configured tmuxp workspaces")

    def get_icon_name(self):
        return "terminal"

    def provides(self):
        yield TmuxSession



__kupfer_actions__ = ("SaveToFile",)

import os
import typing as ty

from kupfer.core import execfile
from kupfer.obj import Action, FileLeaf, TextLeaf, TextSource
from kupfer.obj.compose import ComposedLeaf

if ty.TYPE_CHECKING:
    from gettext import gettext as _


class SaveToFile(Action):
    def __init__(self):
        Action.__init__(self, _("Save As..."))

    def has_result(self):
        return True

    def activate(self, leaf, iobj=None, ctx=None):
        assert iobj
        filepath = iobj.object
        execfile.save_to_file(leaf, filepath)
        execfile.update_icon(leaf, iobj.object)
        return FileLeaf(os.path.abspath(filepath))

    def item_types(self):
        yield ComposedLeaf

    def requires_object(self):
        return True

    def object_types(self):
        yield TextLeaf

    def object_source(self, for_item=None):
        return NameSource(_("Save As..."), ".kfcom")


class NameSource(TextSource):
    """A source for new names for a file;
    here we "autopropose" the source file's extension,
    but allow overriding it as well as renaming to without
    extension (selecting the normal TextSource-returned string).
    """

    def __init__(self, name, extension, sourcefile=None):
        TextSource.__init__(self, name)
        self.sourcefile = sourcefile
        self.extension = extension

    def get_rank(self):
        return 100

    def get_items(self, text):
        if not text:
            return

        t_root, t_ext = os.path.splitext(text)
        yield TextLeaf(text) if t_ext else TextLeaf(t_root + self.extension)

    def get_gicon(self):
        return self.sourcefile and self.sourcefile.get_gicon()

    def get_icon_name(self):
        return "text-x-generic"



"""Pane objects definition.

This file is a part of the program kupfer, which is
released under GNU General Public License v3 (or any later version),
see the main program file, and COPYING for details.
"""

from __future__ import annotations

import itertools
import typing as ty

from gi.repository import GObject

from kupfer.core import actioncompat
from kupfer.core.searcher import Searcher
from kupfer.core.sources import get_source_controller
from kupfer.obj import objects
from kupfer.obj.base import Action, AnySource, KupferObject, Leaf, Source
from kupfer.support import pretty

if ty.TYPE_CHECKING:
    from kupfer.core.search import Rankable

__all__ = (
    "LeafPane",
    "Pane",
    "PrimaryActionPane",
    "SearchContext",
    "SecondaryObjectPane",
)

SearchContext = tuple[int, ty.Any]


def _dress_leaves(
    seq: ty.Iterable[Rankable], action: Action | None
) -> ty.Iterable[Rankable]:
    """yield items of @seq "dressed" by the source controller"""
    sctr = get_source_controller()
    decorate_object = sctr.decorate_object
    for itm in seq:
        decorate_object(itm.object, action=action)  # type:ignore
        yield itm


# Pane Object type definition
PO = ty.TypeVar("PO", bound=KupferObject)


class Pane(GObject.GObject, ty.Generic[PO]):  # type:ignore
    """Pane with `PO` type objects.

    signals:
        search-result (match, match_iter, context)
    """

    __gtype_name__ = "Pane"

    def __init__(self):
        super().__init__()
        self._selection: PO | None = None
        self._latest_key: str | None = None
        self.outstanding_search: int = -1
        self.outstanding_search_id: int = -1
        self._searcher = Searcher()

    def select(self, item: PO | None) -> None:
        self._selection = item

    def get_selection(self) -> PO | None:
        return self._selection

    def reset(self) -> None:
        self._selection = None
        self._latest_key = None

    def get_latest_key(self) -> str | None:
        return self._latest_key

    def get_can_enter_text_mode(self) -> bool:
        return False

    def get_should_enter_text_mode(self) -> bool:
        return False

    def emit_search_result(
        self,
        match: Rankable | None,
        match_iter: ty.Iterable[Rankable],
        context: SearchContext | None,
    ) -> None:
        self.emit("search-result", match, match_iter, context)


GObject.signal_new(
    "search-result",
    Pane,
    GObject.SignalFlags.RUN_LAST,
    GObject.TYPE_BOOLEAN,
    (GObject.TYPE_PYOBJECT, GObject.TYPE_PYOBJECT, GObject.TYPE_PYOBJECT),
)


class LeafPane(Pane[Leaf], pretty.OutputMixin):
    __gtype_name__ = "LeafPane"

    def __init__(self):
        super().__init__()
        # source_stack keep track on history selected sources and leaves
        self._source_stack: list[tuple[AnySource, Leaf | None]] = []
        self._source: AnySource | None = None
        # object_stack is used by "comma trick"
        self.object_stack: list[Leaf] = []

    def select(self, item: Leaf | None) -> None:
        assert item is None or isinstance(item, Leaf), (
            "New selection for object pane is not a Leaf!"
        )
        super().select(item)

    def _load_source(self, src: AnySource) -> AnySource:
        """Try to get a source from the SourceController,
        if it is already loaded we get it from there, else
        returns @src"""
        sctr = get_source_controller()
        return sctr.get_canonical_source(src)

    def get_source(self) -> AnySource | None:
        return self._source

    def source_rebase(self, src: AnySource) -> None:
        self._source_stack.clear()
        self._source = self._load_source(src)
        self.refresh_data()

    def push_source(self, src: AnySource) -> None:
        if self._source:
            self._source_stack.append((self._source, self._selection))

        self._source = self._load_source(src)
        self.refresh_data()

    def _pop_source(self) -> bool:
        """Remove source from stack. Return True if succeeded"""
        if self._source_stack:
            self._source, self._selection = self._source_stack.pop()
            return True

        return False

    def is_at_source_root(self) -> bool:
        """Return True if we have no source stack"""
        return not self._source_stack

    def object_stack_push(self, obj: Leaf) -> None:
        self.object_stack.append(obj)

    def object_stack_pop(self) -> Leaf:
        return self.object_stack.pop()

    def get_can_enter_text_mode(self) -> bool:
        return self.is_at_source_root()

    def get_should_enter_text_mode(self) -> bool:
        return False

    def refresh_data(self, select: ty.Any = None) -> None:
        self.emit("new-source", self._source, select)

    def browse_up(self) -> bool:
        """Try to browse up to previous sources, from current source"""
        succ = self._pop_source()
        if not succ:
            assert self._source
            if self._source.has_parent():
                self.source_rebase(self._source.get_parent())  # type:ignore
                succ = True

        if succ:
            sel = self._selection
            # select again only leaves that have some content (as we go from
            # source we can back only to source)
            if sel and hasattr(sel, "has_content") and sel.has_content():
                self.refresh_data(select=sel)
            else:
                self.refresh_data()

        return succ

    def browse_down(self, alternate: bool = False) -> bool:
        """Browse into @leaf if it's possible and save away the previous sources
        in the stack. If @alternate, use the Source's alternate method."""
        leaf: Leaf | None = self.get_selection()
        if (
            leaf
            and leaf.has_content()
            and (csrc := leaf.content_source(alternate=alternate))
        ):
            self.push_source(csrc)
            return True

        return False

    def reset(self) -> None:
        """Pop all sources and go back to top level"""
        Pane.reset(self)
        while self._pop_source():
            pass

        self.refresh_data()

    def soft_reset(self) -> AnySource | None:
        Pane.reset(self)
        while self._pop_source():
            pass

        return self._source

    def search(
        self,
        key: str = "",
        context: SearchContext | None = None,
        text_mode: bool = False,
    ) -> None:
        """Search for `key`"""

        self._latest_key = key
        sources_: ty.Iterable[AnySource] = ()
        if not text_mode and (srcs := self.get_source()):
            sources_ = (srcs,)

        if key and self.is_at_source_root():
            # Only use text sources when we are at root catalog
            sctr = get_source_controller()
            textsrcs = sctr.get_text_sources()
            sources_ = itertools.chain(sources_, textsrcs)

        def _decorator(seq):
            return _dress_leaves(seq, action=None)

        match, match_iter = self._searcher.search(
            sources_, key, score=bool(key), decorator=_decorator
        )
        self.emit_search_result(match, match_iter, context)


GObject.signal_new(
    "new-source",
    LeafPane,
    GObject.SignalFlags.RUN_LAST,
    GObject.TYPE_BOOLEAN,
    (GObject.TYPE_PYOBJECT, GObject.TYPE_PYOBJECT),
)


class PrimaryActionPane(Pane[Action]):
    def __init__(self):
        super().__init__()
        self._action_valid_cache: dict[int, bool] = {}
        self.set_item(None)

    def select(self, item: Action | None) -> None:
        assert not item or isinstance(item, Action), (
            "Selection in action pane is not an Action!"
        )
        super().select(item)

    def set_item(self, item: Leaf | None) -> None:
        """Set which @item we are currently listing actions for"""
        self._current_item = item
        self._action_valid_cache.clear()

    def search(
        self,
        key: str = "",
        context: SearchContext | None = None,
        text_mode: bool = False,
    ) -> None:
        """Search: Register the search method in the event loop using @key,
        promising to return @context in the notification about the result,
        having selected @item in PaneSel.SOURCE

        If we already have a call to search, we remove the "source"
        so that we always use the most recently requested search."""

        self._latest_key = key
        leaf = self._current_item
        if not leaf:
            self.emit_search_result(None, (), context)
            return

        if isinstance(leaf, objects.ActionLeaf):
            # for ActionLeaf get only actions defined in leaf (should be one,
            # ignore other actions).
            actions = leaf.get_actions()
            match, match_iter = self._searcher.rank_actions(actions, "", leaf)

        else:
            actions = actioncompat.actions_for_item(
                leaf, get_source_controller()
            )
            cache = self._action_valid_cache

            def valid_decorator(seq):
                """Check if actions are valid before access"""
                assert leaf

                for obj in seq:
                    action = obj.object
                    action_hash = hash(action)
                    valid = cache.get(action_hash)
                    if valid is None:
                        valid = actioncompat.action_valid_for_item(
                            action, leaf
                        )
                        cache[action_hash] = valid

                    if valid:
                        yield obj

            match, match_iter = self._searcher.rank_actions(
                actions, key, leaf, decorator=valid_decorator
            )

        self.emit_search_result(match, match_iter, context)


class SecondaryObjectPane(LeafPane):
    __gtype_name__ = "SecondaryObjectPane"

    def __init__(self):
        LeafPane.__init__(self)
        self._current_item: Leaf | None = None
        self._current_action: Action | None = None

    def reset(self) -> None:
        LeafPane.reset(self)
        self._searcher.reset()

    def set_item_and_action(
        self, item: Leaf | None, act: Action | None
    ) -> None:
        self._current_item = item
        self._current_action = act
        if item and act:
            ownsrc, use_catalog = actioncompat.iobject_source_for_action(
                act, item
            )
            if ownsrc and not use_catalog:
                self.source_rebase(ownsrc)
            else:
                extra_sources = (
                    ty.cast("list[Source]", [ownsrc]) if ownsrc else None
                )

                sctr = get_source_controller()
                self.source_rebase(
                    sctr.root_for_types(act.object_types(), extra_sources)
                )
        else:
            self.reset()

    def get_can_enter_text_mode(self) -> bool:
        """Check if there are any reasonable text sources for this action"""
        assert self._current_action
        atroot = self.is_at_source_root()
        types = tuple(self._current_action.object_types())
        sctr = get_source_controller()
        textsrcs = sctr.get_text_sources()
        return atroot and any(
            sctr.good_source_for_types(s, types) for s in textsrcs
        )

    def get_should_enter_text_mode(self):
        return self.is_at_source_root() and hasattr(
            self.get_source(), "get_text_items"
        )

    def search(
        self,
        key: str = "",
        context: SearchContext | None = None,
        text_mode: bool = False,
    ) -> None:
        """
        filter for action @item
        """
        self._latest_key = key

        assert self._current_action
        if not self._current_item:
            self.emit_search_result(None, (), context)
            return

        sources_: ty.Iterable[AnySource] = []
        if (
            not text_mode or hasattr(self.get_source(), "get_text_items")
        ) and (srcs := self.get_source()):
            sources_ = itertools.chain(sources_, (srcs,))

        if key and self.is_at_source_root():
            # Only use text sources when we are at root catalog
            sctr = get_source_controller()
            if textsrcs := sctr.get_text_sources():
                sources_ = itertools.chain(sources_, textsrcs)

        item_check = actioncompat.iobjects_valid_for_action(
            self._current_action, self._current_item
        )

        def decorator(seq):
            return _dress_leaves(seq, action=self._current_action)

        match, match_iter = self._searcher.search(
            sources_,
            key,
            score=True,
            item_check=item_check,
            decorator=decorator,
        )
        self.emit_search_result(match, match_iter, context)



__kupfer_name__ = _("Pinboard")
__kupfer_sources__ = ("PinboardBookmarkSource",)
__kupfer_actions__ = ()
__description__ = _("Manage and use bookmarks from Pinboard")
__version__ = "2020-11-15"
__author__ = "Peter Stuifzand <peter@p83.nl>"

import typing as ty

import pinboard  # pylint: disable=import-error

from kupfer import plugin_support
from kupfer.obj import Source, UrlLeaf

if ty.TYPE_CHECKING:
    from gettext import gettext as _

__kupfer_settings__ = plugin_support.PluginSettings(
    {
        "key": "token",
        "label": _("Pinboard API Token"),
        "type": str,
        "value": "",
    },
)


class PinboardBookmarkSource(Source):
    def __init__(self):
        super().__init__(_("Pinboard Bookmarks"))

    def get_items(self):
        token = __kupfer_settings__["token"]
        if token == "":
            return []

        pinb = pinboard.Pinboard(token)
        bookmarks = pinb.posts.all(start=0, results=100)
        return [UrlLeaf(b.url, b.description) for b in bookmarks]

    def get_description(self):
        return _("Index of Pinboard bookmarks")

    def get_gicon(self):
        if lrepr := self.get_leaf_repr():
            return lrepr.get_gicon()

        return None

    def get_icon_name(self):
        return "web-browser"

    def provides(self):
        yield UrlLeaf



from __future__ import annotations

__kupfer_name__ = _("APT")
__kupfer_sources__ = ()
__kupfer_text_sources__ = ()
__kupfer_actions__ = (
    "ShowPackageInfo",
    "SearchPackageName",
    "InstallPackage",
    "OpenPackageWebsite",
    "SearchForFile",
)
__description__ = _("Interface with the package manager APT")
__version__ = ""
__author__ = (
    "Martin Koelewijn <martinkoelewijn@gmail.com>, "
    "Ulrik Sverdrup <ulrik.sverdrup@gmail.com>"
)

import subprocess
import typing as ty
import urllib.parse

from kupfer import icons, launch, plugin_support, support
from kupfer.obj import Action, Leaf, Source, TextLeaf
from kupfer.support import kupferstring, task
from kupfer.ui import uiutils

if ty.TYPE_CHECKING:
    from gettext import gettext as _

__kupfer_settings__ = plugin_support.PluginSettings(
    {
        "key": "installation_method",
        "label": _("Installation method"),
        "type": str,
        "value": "gksu -- apt-get install --yes",
    },
)

plugin_support.check_command_available("apt", "apt-cache")


# pylint: disable=too-few-public-methods
class InfoTask(task.Task):
    def __init__(self, text: str):
        super().__init__()
        self.text = text
        self.aptitude: bytes | None = None
        self.apt_cache: bytes | None = None
        self._finish_callback = None

    def start(self, finish_callback):
        self._finish_callback = finish_callback
        timeout = 60
        launch.AsyncCommand(
            ["apt", "show", self.text], self._aptitude_finished, timeout
        )
        launch.AsyncCommand(
            ["apt-cache", "policy", self.text],
            self._aptcache_finished,
            timeout,
        )

    def _aptitude_finished(
        self, acommand: launch.AsyncCommand, stdout: bytes, stderr: bytes
    ) -> None:
        self.aptitude = stderr + stdout
        self._check_end()

    def _aptcache_finished(
        self, acommand: launch.AsyncCommand, stdout: bytes, stderr: bytes
    ) -> None:
        self.apt_cache = stderr + stdout
        self._check_end()

    def _check_end(self):
        if self.aptitude is None or self.apt_cache is None:
            return

        assert self._finish_callback
        text = kupferstring.fromlocale(
            self.aptitude
        ) + kupferstring.fromlocale(self.apt_cache)
        uiutils.show_text_result(text, title=_("Show Package Information"))
        self._finish_callback(self)


class ShowPackageInfo(Action):
    def __init__(self):
        Action.__init__(self, _("Show Package Information"))
        self.kupfer_add_alias("apt show")

    def is_async(self):
        return True

    def activate(self, leaf, iobj=None, ctx=None):
        return InfoTask(leaf.object.strip())

    def item_types(self):
        yield TextLeaf
        yield Package

    def valid_for_item(self, leaf):
        # check if it is a single word
        text = leaf.object
        return len(text.split(None, 1)) == 1

    def get_gicon(self):
        return icons.ComposedIcon("dialog-information", "package-x-generic")


class OpenPackageWebsite(Action):
    def __init__(self):
        Action.__init__(self, _("Browse packages.debian.org"))

    def activate(self, leaf, iobj=None, ctx=None):
        name = leaf.object.strip()
        url = f"https://packages.debian.org/{name}"
        launch.show_url(url)

    def get_description(self):
        return _("Open packages.debian.org page for package")

    def item_types(self):
        yield TextLeaf
        yield Package

    def valid_for_item(self, leaf):
        # check if it is a single word
        text = leaf.object
        return len(text.split(None, 1)) == 1

    def get_gicon(self):
        return icons.ComposedIcon("edit-find", "package-x-generic")


class InstallPackage(Action):
    def __init__(self):
        Action.__init__(self, _("Install"))
        self.kupfer_add_alias("apt install")

    def activate(self, leaf, iobj=None, ctx=None):
        self.activate_multiple((leaf,))

    def activate_multiple(self, objs):
        program = __kupfer_settings__["installation_method"]
        pkgs = [o.object.strip() for o in objs]
        prog_argv = support.argv_for_commandline(program)
        launch.spawn_in_terminal(prog_argv + pkgs)

    def item_types(self):
        yield Package
        yield TextLeaf

    def get_description(self):
        return _("Install package using the configured method")

    def get_icon_name(self):
        return "document-save"


class Package(Leaf):
    def __init__(self, package: str, desc: str):
        Leaf.__init__(self, package, package)
        self.desc = desc

    def get_text_representation(self):
        return self.object

    def get_description(self):
        return self.desc

    def get_icon_name(self):
        return "package-x-generic"


class PackageSearchSource(Source):
    def __init__(self, query):
        self.query = query
        Source.__init__(self, _('Packages matching "%s"') % query)

    def repr_key(self):
        return self.query

    def get_items(self):
        proc = subprocess.run(
            ["apt-cache", "search", "--names-only", self.query],
            capture_output=True,
            check=True,
        )
        for line in kupferstring.fromlocale(proc.stdout).splitlines():
            line = line.strip()  # noqa: PLW2901
            if not line:
                continue

            package, sep, desc = line.partition(" - ")
            if not sep:
                self.output_error("apt-cache: ", line)
                continue

            yield Package(package, desc)

    def should_sort_lexically(self):
        return True

    def provides(self):
        yield TextLeaf

    def get_icon_name(self):
        return "system-software-install"


class SearchPackageName(Action):
    def __init__(self):
        Action.__init__(self, _("Search Package Name..."))
        self.kupfer_add_alias("apt search")

    def is_factory(self):
        return True

    def activate(self, leaf, iobj=None, ctx=None):
        package = leaf.object.strip()
        return PackageSearchSource(package)

    def item_types(self):
        yield TextLeaf

    def valid_for_item(self, leaf):
        # check if it is a single word
        text = leaf.object
        return len(text.split(None, 1)) == 1

    def get_icon_name(self):
        return "system-software-install"


class SearchForFile(Action):
    def __init__(self):
        Action.__init__(self, _("Search for file in packages..."))
        self.kupfer_add_alias("apt file")

    def activate(self, leaf, iobj=None, ctx=None):
        name = urllib.parse.quote(leaf.object.strip())
        url = f"https://packages.debian.org/file:{name}"
        launch.show_url(url)

    def get_description(self):
        return _(
            "Search the contents of Debian distributions for any files (online)"
        )

    def item_types(self):
        yield TextLeaf

    def valid_for_item(self, leaf):
        text = leaf.object
        return len(text) > 2  # noqa: PLR2004

    def get_gicon(self):
        return icons.ComposedIcon("edit-find", "package-x-generic")



__kupfer_name__ = _("Kupfer Plugins")
__kupfer_sources__ = ("KupferPlugins",)
__description__ = _("Access Kupfer's plugin list in Kupfer")
__version__ = ""
__author__ = "Ulrik Sverdrup <ulrik.sverdrup@gmail.com>"

import importlib.util
import os
import typing as ty
from pathlib import Path

# Since this is a core plugin we break some rules
# These modules are normally out of bounds for plugins
from kupfer.core import plugins, settings
from kupfer.obj import Action, FileLeaf, Leaf, Source, TextLeaf
from kupfer.ui import preferences

if ty.TYPE_CHECKING:
    from gettext import gettext as _


class ShowInfo(Action):
    def __init__(self):
        Action.__init__(self, _("Show Information"))

    def wants_context(self):
        return True

    def activate(self, leaf, iobj=None, ctx=None):
        assert ctx
        plugin_id = leaf.object["name"]
        preferences.show_plugin_info(plugin_id, ctx.environment)

    def get_description(self):
        pass

    def get_icon_name(self):
        return "dialog-information"


class ShowSource(Action):
    def __init__(self):
        Action.__init__(self, _("Show Source Code"))

    def has_result(self):
        return True

    def activate(self, leaf, iobj=None, ctx=None):
        # Try to find the __file__ attribute for the plugin
        # It will fail for files inside zip packages, but that is
        # uncommon for now.
        # Additionally, it will fail for fake plugins
        plugin_id = leaf.object["name"]
        filename: str = plugins.get_plugin_attribute(  # type:ignore
            plugin_id, plugins.PluginAttr.FILE
        )
        if not filename:
            return leaf

        root, ext = os.path.splitext(filename)
        if ext.lower() == ".pyc" and Path(root + ".py").exists():
            return FileLeaf(root + ".py")

        if not Path(filename).exists():
            # handle modules in zip or eggs
            pfull = "kupfer.plugin." + plugin_id
            try:
                spec = importlib.util.find_spec(pfull)
            except (ImportError, AttributeError, TypeError, ValueError):
                pass
            else:
                loader = spec.loader if spec is not None else None
                if loader:
                    return TextLeaf(loader.get_source(pfull))  # type:ignore

        return FileLeaf(filename)

    def get_description(self):
        pass

    def get_icon_name(self):
        return "dialog-information"


class Plugin(Leaf):
    # NOTE: Just to be sure that a plugin ranks lower than a
    # like-named other object by default.
    rank_adjust = -1

    def __init__(self, obj, name):
        Leaf.__init__(self, obj, name)

    def get_actions(self):
        yield ShowInfo()
        yield ShowSource()

    def get_description(self):
        setctl = settings.get_settings_controller()
        enabled = setctl.get_plugin_enabled(self.object["name"])
        enabled_str = _("enabled") if enabled else _("disabled")
        return f"{self.object['description']} ({enabled_str})"

    def get_icon_name(self):
        return "package-x-generic"


class KupferPlugins(Source):
    def __init__(self):
        Source.__init__(self, _("Kupfer Plugins"))

    def get_items(self):
        setctl = settings.get_settings_controller()
        for info in plugins.get_plugin_info():
            plugin_id = info["name"]
            if setctl.get_plugin_is_hidden(plugin_id):
                continue

            yield Plugin(info, info["localized_name"])

    def should_sort_lexically(self):
        return True

    def provides(self):
        yield Plugin

    def get_icon_name(self):
        return "package-x-generic"

    @classmethod
    def decorates_type(cls):
        return Plugin

    @classmethod
    def decorate_item(cls, obj):
        if cls.is_self_plugin(obj):
            return cls()

        return None

    @classmethod
    def is_self_plugin(cls, obj):
        self_plug_id = __name__.rsplit(".", maxsplit=1)[-1]
        return obj.object["name"] == self_plug_id

    def get_leaf_repr(self):
        for obj in self.get_leaves() or []:
            if self.is_self_plugin(obj):
                return obj

        return None



"""
Work with Textfiles: Allow appending and writing new files,
or extracting the content of files.

All Text in Kupfer is in unicode. When we read from textfiles or write
to textfiles, we always work in the locale-defined encoding.

FIXME: Be less strict (use UTF-8 if locale says Ascii)
"""

__kupfer_name__ = _("Textfiles")
__kupfer_actions__ = (
    "AppendTo",
    "AppendText",
    "WriteTo",
    "GetTextContents",
    "CopyContent",
)
__description__ = _("Action for text files")
__version__ = "2017.1"
__author__ = ""

import typing as ty
from pathlib import Path

from gi.repository import Gdk, Gtk

from kupfer.obj import Action, FileLeaf, TextLeaf, helplib
from kupfer.support import fileutils, kupferstring, validators

if ty.TYPE_CHECKING:
    from gettext import gettext as _


class AppendTo(Action):
    def __init__(self, name=None):
        super().__init__(name or _("Append To..."))

    def activate(self, leaf, iobj=None, ctx=None):
        with open(
            iobj.object, "a", encoding=kupferstring.get_encoding()
        ) as outfile:
            outfile.write(leaf.object)
            outfile.write("\n")

    def item_types(self):
        yield TextLeaf

    def requires_object(self):
        return True

    def object_types(self):
        yield FileLeaf

    def valid_object(self, iobj, for_item=None):
        # K: allow select all writable FileLeaves; filtering by content
        # prevent navigate between directories
        return iobj.is_writable()
        # return iobj.is_content_type("text/plain")

    def get_icon_name(self):
        return "list-add"


class AppendText(helplib.reverse_action(AppendTo)):  # type: ignore
    def __init__(self):
        super().__init__(_("Append..."))


class WriteTo(Action):
    def __init__(self):
        super().__init__(_("Write To..."))

    def has_result(self):
        return True

    def activate(self, leaf, iobj=None, ctx=None):
        assert iobj

        if isinstance(iobj, TextLeaf):
            outfile, outpath = fileutils.get_destfile(iobj.object)
        elif isinstance(iobj, FileLeaf):
            outfile, outpath = fileutils.get_destfile_in_directory(
                iobj.object, _("Empty File")
            )
        else:
            # TODO: raise TypeError
            raise ValueError  # noqa:TRY004

        if not outfile or not outpath:
            return None

        try:
            text = str(leaf.object).encode()
            outfile.write(text)
            if not text.endswith(b"\n"):
                outfile.write(b"\n")
        finally:
            outfile.close()

        return FileLeaf(outpath)

    def item_types(self):
        yield TextLeaf

    def requires_object(self):
        return True

    def object_types(self):
        yield FileLeaf
        yield TextLeaf

    def valid_object(self, iobj, for_item=None):
        if isinstance(iobj, FileLeaf):
            return iobj.is_dir()

        # we accept TextLeaf if it look like path
        path_str = str(iobj.object)
        if not validators.is_valid_file_path(path_str):
            return False

        path = Path(path_str)
        # file should not exist
        if path.exists():
            return False

        # but parent dir must exists and be writable
        return fileutils.is_directory_writable(path.parent)

    def get_description(self):
        return _("Write the text to a new file in specified directory")

    def get_icon_name(self):
        return "document-new"


class GetTextContents(Action):
    def __init__(self):
        super().__init__(_("Get Text Contents"))

    def has_result(self):
        return True

    def activate(self, leaf, iobj=None, ctx=None):
        text = Path(leaf.object).read_text(
            encoding=kupferstring.get_encoding()
        )
        return TextLeaf(text)

    def item_types(self):
        yield FileLeaf

    def valid_for_item(self, leaf):
        return leaf.is_content_type("text/plain")

    def get_icon_name(self):
        return "edit-copy"


class CopyContent(Action):
    action_accelerator = "c"

    def __init__(self):
        super().__init__(_("Copy content"))

    def has_result(self):
        return True

    def item_types(self):
        yield FileLeaf

    def valid_for_item(self, leaf):
        return leaf.is_content_type("text/plain")

    def get_icon_name(self):
        return "edit-copy"

    def wants_context(self):
        return True

    def activate(self, leaf, iobj=None, ctx=None):
        assert ctx
        self.activate_multiple((leaf,), ctx=ctx)

    def activate_multiple(self, objs, iobjects=None, ctx=None):
        assert ctx

        clip = Gtk.Clipboard.get_for_display(
            ctx.environment.get_screen().get_display(), Gdk.SELECTION_CLIPBOARD
        )

        content = []
        for obj in objs:
            try:
                content.append(
                    Path(obj.object).read_text(
                        encoding=kupferstring.get_encoding()
                    )
                )
            except Exception as err:  # noqa: PERF203
                content.append(f"Read file {obj.object} error: {err}")

        clip.set_text("\n".join(content), -1)  # -1 for computed string length

    def get_description(self):
        return _("Copy file content to clipboard")



"""
File-related objects

This file is a part of the program kupfer, which is
released under GNU General Public License v3 (or any later version),
see the main program file, and COPYING for details.
"""

from __future__ import annotations

import os
import typing as ty
from os import path
from pathlib import Path

from gi.repository import GdkPixbuf, Gio, GLib

from kupfer import icons, launch
from kupfer.obj import actions, fileactions, filesrc
from kupfer.obj.base import Action, Leaf, Source
from kupfer.obj.exceptions import InvalidDataError
from kupfer.obj.representation import TextRepresentation
from kupfer.support import pretty

__all__ = ("FileLeaf",)


class FileLeaf(Leaf, TextRepresentation):
    """Represents one file: the represented object is a string."""

    serializable: int | None = 1

    def __init__(
        self,
        obj: str | Path,
        name: str | None = None,
        alias: str | None = None,
    ) -> None:
        """Construct a FileLeaf

        The display name of the file is normally derived from the full path,
        and @name should normally be left unspecified.

        @obj: string (path) or Path object
        @name: file name or None for using basename
        """
        if obj is None:
            raise InvalidDataError(f"File path for {name} may not be None")

        if isinstance(obj, Path):
            obj = str(obj)

        # Use glib filename reading to make display name out of filenames
        # this function returns a `unicode` object
        if not name:
            name = GLib.filename_display_basename(obj)

        assert name
        super().__init__(obj, name)
        if alias:
            self.kupfer_add_alias(alias)

    @classmethod
    def from_uri(cls, uri: str) -> FileLeaf | None:
        """Construct a FileLeaf from local `uri`.
        Return FileLeaf if it is supported, else None.
        """
        gfile = Gio.File.new_for_uri(uri)
        fpath = gfile.get_path()
        if fpath:
            return cls(fpath)

        return None

    def __hash__(self) -> int:
        return hash(str(self))

    def __eq__(self, other: object) -> bool:
        try:
            return (
                type(self) is type(other)
                and str(self) == str(other)
                and path.samefile(self.object, other.object)  # type: ignore
            )
        except OSError as exc:
            pretty.print_debug(__name__, exc)
            return False

    def repr_key(self) -> ty.Any:
        return self.object

    def canonical_path(self) -> str:
        """Return the true path of the File (without symlinks)"""
        assert isinstance(self.object, str)
        return path.realpath(self.object)

    def is_valid(self) -> bool:
        return os.access(self.object, os.R_OK)

    def is_writable(self) -> bool:
        return os.access(self.object, os.W_OK)

    def is_executable(self) -> bool:
        return not path.isdir(self.object) and os.access(
            self.object, os.R_OK | os.X_OK
        )

    def is_dir(self) -> bool:
        return path.isdir(self.object)

    def get_text_representation(self) -> str:
        return GLib.filename_display_name(self.object)  # type: ignore

    def get_urilist_representation(self) -> list[str]:
        return [self.get_gfile().get_uri()]

    def get_gfile(self) -> Gio.File:
        """Return a Gio.File of self."""
        return Gio.File.new_for_path(self.object)

    def get_description(self) -> str | None:
        return launch.get_display_path_for_bytestring(self.canonical_path())

    def get_actions(self) -> ty.Iterable[Action]:
        yield fileactions.Open()
        yield fileactions.GetParent()

        if self.is_dir():
            yield actions.OpenTerminal()

        elif self.is_valid() and self._is_good_executable():
            yield actions.Execute()
            yield actions.Execute(in_terminal=True)

    def has_content(self) -> bool:
        return self.is_dir() or Leaf.has_content(self)

    def content_source(self, alternate: bool = False) -> Source | None:
        if self.is_dir():
            return filesrc.DirectorySource(self.object, show_hidden=alternate)

        return Leaf.content_source(self)

    def get_thumbnail(
        self, width: int, height: int
    ) -> GdkPixbuf.Pixbuf | None:
        if self.is_dir():
            return None

        return icons.get_thumbnail_for_gfile(self.get_gfile(), width, height)

    def get_gicon(self) -> icons.GIcon | None:
        return icons.get_gicon_for_file(self.object)

    def get_icon_name(self) -> str:
        if self.is_dir():
            return "folder"

        return "text-x-generic"

    def get_content_type(self) -> str | None:
        ret: str
        uncertain: bool
        ret, uncertain = Gio.content_type_guess(self.object, None)
        if not uncertain:
            return ret

        content_attr = Gio.FILE_ATTRIBUTE_STANDARD_CONTENT_TYPE
        gfile = self.get_gfile()
        if not gfile.query_exists(None):
            return None

        info = gfile.query_info(
            content_attr, Gio.FileQueryInfoFlags.NONE, None
        )
        content_type: str = info.get_attribute_string(content_attr)
        return content_type

    def is_content_type(self, ctype: str) -> bool:
        """Return True if this file is of the type `ctype` mime type, can have
        wildcards like 'image/*'."""
        predicate = Gio.content_type_is_a
        ctype_guess, uncertain = Gio.content_type_guess(self.object, None)
        ret: bool = predicate(ctype_guess, ctype)
        if ret or not uncertain:
            return ret

        content_attr = Gio.FILE_ATTRIBUTE_STANDARD_CONTENT_TYPE
        gfile = self.get_gfile()
        if not gfile.query_exists(None):
            return False

        info = gfile.query_info(
            content_attr, Gio.FileQueryInfoFlags.NONE, None
        )
        content_type = info.get_attribute_string(content_attr)
        return predicate(content_type, ctype)  # type: ignore

    def _is_good_executable(self):
        if not self.is_executable():
            return False

        ctype, uncertain = Gio.content_type_guess(self.object, None)
        return uncertain or Gio.content_type_can_be_executable(ctype)



__kupfer_name__ = _("Nemo")
__kupfer_sources__ = ("FmObjects",)
__kupfer_actions__ = ("Reveal", "GetInfo", "CopyTo")
__description__ = _("File manager actions")
__version__ = "2017.1"
__author__ = ""

import os
import typing as ty

import dbus

from kupfer import plugin_support
from kupfer.obj import Action, FileLeaf, OperationError, RunnableLeaf, Source
from kupfer.obj.apps import AppLeafContentMixin
from kupfer.support import pretty

if ty.TYPE_CHECKING:
    from gettext import gettext as _

plugin_support.check_dbus_connection()

_SERVICE_NAME = "org.Nemo"
_FO_OBJECT = "/org/Nemo"
_FO_IFACE = "org.Nemo.FileOperations"

_FM_OBJECT = "/org/freedesktop/FileManager1"
_FM_IFACE = "org.freedesktop.FileManager1"


def _get_fm1():
    bus = dbus.SessionBus()
    try:
        proxy_obj = bus.get_object(_SERVICE_NAME, _FM_OBJECT)
    except dbus.DBusException as exc:
        raise OperationError(exc) from exc

    return dbus.Interface(proxy_obj, _FM_IFACE)


def _get_nemo():
    bus = dbus.SessionBus()
    try:
        proxy_obj = bus.get_object(_SERVICE_NAME, _FO_OBJECT)
    except dbus.DBusException as exc:
        raise OperationError(exc) from exc

    return dbus.Interface(proxy_obj, _FO_IFACE)


def _dummy(*args):
    pass


def make_error_handler(ctx):
    def error_handler(exc):
        ctx.register_late_error(exc)

    return error_handler


class Reveal(Action):
    def __init__(self):
        Action.__init__(self, _("Select in File Manager"))

    def wants_context(self):
        return True

    def activate(self, leaf, iobj=None, ctx=None):
        assert ctx
        return self.activate_multiple((leaf,), ctx)

    def activate_multiple(self, leaves, ctx):
        uris = [leaf_uri(leaf) for leaf in leaves]
        id_ = ctx.environment.get_startup_notification_id()
        _get_fm1().ShowItems(
            uris,
            id_,
            reply_handler=_dummy,
            error_handler=make_error_handler(ctx),
        )

    def item_types(self):
        yield FileLeaf


class GetInfo(Action):
    def __init__(self):
        Action.__init__(self, _("Show Properties"))

    def wants_context(self):
        return True

    def activate(self, leaf, iobj=None, ctx=None):
        assert ctx
        return self.activate_multiple((leaf,), ctx)

    def activate_multiple(self, leaves, ctx):
        uris = [leaf_uri(leaf) for leaf in leaves]
        id_ = ctx.environment.get_startup_notification_id()
        _get_fm1().ShowItemProperties(
            uris,
            id_,
            reply_handler=_dummy,
            error_handler=make_error_handler(ctx),
        )

    def item_types(self):
        yield FileLeaf

    def get_description(self):
        return _("Show information about file in file manager")

    def get_icon_name(self):
        return "dialog-information"


def _good_destination(dpath, spath):
    """If directory path @dpath is a valid destination for file @spath
    to be copied or moved to.
    """
    if not os.path.isdir(dpath):
        return False

    spath = os.path.normpath(spath)
    dpath = os.path.normpath(dpath)
    cpfx = os.path.commonprefix((spath, dpath))
    return not (os.path.samefile(dpath, spath) or cpfx == spath)


def leaf_uri(leaf):
    return leaf.get_gfile().get_uri()


class CopyTo(Action, pretty.OutputMixin):
    def __init__(self):
        Action.__init__(self, _("Copy To..."))

    def wants_context(self):
        return True

    def activate(self, leaf, iobj=None, ctx=None):
        assert iobj
        assert ctx
        return self.activate_multiple((leaf,), (iobj,), ctx)

    def activate_multiple(self, leaves, iobjects, ctx):
        # Unroll by looping over the destinations,
        # copying everything into each destination
        nemo = _get_nemo()
        source_uris = [leaf_uri(L) for L in leaves]

        def _reply(*args):
            self.output_debug("reply got for copying", *args)

        for dest_iobj in iobjects:
            desturi = leaf_uri(dest_iobj)
            nemo.CopyURIs(
                source_uris,
                desturi,
                reply_handler=_reply,
                error_handler=make_error_handler(ctx),
            )

    def item_types(self):
        yield FileLeaf

    def valid_for_item(self, leaf):
        return True

    def requires_object(self):
        return True

    def object_types(self):
        yield FileLeaf

    def valid_object(self, obj, for_item):
        return _good_destination(obj.object, for_item.object)

    def get_description(self):
        return _("Copy file to a chosen location")


class EmptyTrash(RunnableLeaf):
    def __init__(self):
        RunnableLeaf.__init__(self, None, _("Empty Trash"))

    def wants_context(self):
        return True

    def run(self, ctx=None):
        assert ctx
        _get_nemo().EmptyTrash(
            reply_handler=_dummy, error_handler=make_error_handler(ctx)
        )

    def get_description(self):
        return None

    def get_icon_name(self):
        return "user-trash-full"


class FmObjects(AppLeafContentMixin, Source):
    appleaf_content_id = "nemo"

    def __init__(self):
        Source.__init__(self, _("Nemo"))

    def get_items(self):
        yield EmptyTrash()

    def provides(self):
        yield RunnableLeaf

    def get_icon_name(self):
        return "nemo"



"""
File - related sources

This file is a part of the program kupfer, which is
released under GNU General Public License v3 (or any later version),
see the main program file, and COPYING for details.
"""

from __future__ import annotations

import os
import typing as ty
from contextlib import suppress
from os import path

from gi.repository import Gio, GLib

from kupfer import icons
from kupfer.obj import apps, files
from kupfer.obj.base import Leaf, Source
from kupfer.obj.exceptions import InvalidDataError
from kupfer.obj.helplib import FilesystemWatchMixin
from kupfer.support import fileutils

if ty.TYPE_CHECKING:
    from gettext import gettext as _

__all__ = (
    "DirectorySource",
    "FileSource",
    "construct_file_leaf",
)


def construct_file_leaf(obj: str) -> Leaf:
    """If the path in @obj points to a Desktop Item file,
    return an AppLeaf, otherwise return a FileLeaf
    """
    if obj.endswith(".desktop"):
        with suppress(InvalidDataError):
            return apps.AppLeaf(init_path=obj)

    return files.FileLeaf(obj)


class DirectorySource(Source, FilesystemWatchMixin):
    source_use_cache = False

    def __init__(
        self,
        directory: str,
        show_hidden: bool = False,
        *,
        toplevel: bool = False,
    ) -> None:
        # Use glib filename reading to make display name out of filenames
        # this function returns a `unicode` object
        name = GLib.filename_display_basename(directory)
        super().__init__(name)
        self._directory = directory
        self._show_hidden = show_hidden
        self._toplevel = toplevel
        self.monitor: ty.Any = None
        self._user_home = os.path.expanduser("~")

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__module__}.{self.__class__.__name__}"
            f'("{self._directory}", show_hidden={self._show_hidden})'
        )

    def initialize(self) -> None:
        # only toplevel directories are active monitored
        if self._toplevel:
            self.monitor = self.monitor_directories(self._directory)

    def finalize(self) -> None:
        if self.monitor:
            self.stop_monitor_fs_changes(self.monitor)
            self.monitor = None

    def monitor_include_file(self, gfile: Gio.File) -> bool:
        return self._show_hidden or not gfile.get_basename().startswith(".")

    def get_items(self) -> ty.Iterator[Leaf]:
        dirfiles: ty.Iterable[os.DirEntry[str]]
        try:
            with os.scandir(self._directory) as dirfiles:
                if self._show_hidden:
                    yield from (
                        construct_file_leaf(entry.path) for entry in dirfiles
                    )

                else:
                    yield from (
                        construct_file_leaf(entry.path)
                        for entry in dirfiles
                        if entry.name[0] != "."
                    )

        except OSError as exc:
            self.output_error(exc)

    def should_sort_lexically(self) -> bool:
        return True

    def _parent_path(self) -> str:
        return path.normpath(path.join(self._directory, path.pardir))

    def has_parent(self) -> bool:
        return not path.samefile(self._directory, self._parent_path())

    def get_parent(self) -> DirectorySource | None:
        if not self.has_parent():
            return None

        return DirectorySource(self._parent_path())

    def get_description(self) -> str:
        return _("Directory source %s") % self._directory

    def get_gicon(self) -> icons.GIcon | None:
        return icons.get_gicon_for_file(self._directory)

    def get_icon_name(self) -> str:
        return "folder"

    def get_leaf_repr(self) -> Leaf | None:
        alias = None
        directory = self._directory
        if os.path.samefile(directory, self._user_home):
            alias = _("Home Folder")

        return files.FileLeaf(directory, alias=alias)

    def provides(self) -> ty.Iterable[ty.Type[Leaf]]:
        yield files.FileLeaf
        yield apps.AppLeaf


class FileSource(Source):
    source_use_cache = False

    def __init__(self, dirlist: list[str], depth: int = 0) -> None:
        """
        @dirlist: Directories as byte strings
        """
        name = GLib.filename_display_basename(dirlist[0])
        if len(dirlist) > 1:
            name = _("%s et. al.") % name

        super().__init__(name)
        self.dirlist = dirlist
        self.depth = depth

    def __repr__(self) -> str:
        mod = self.__class__.__module__
        cname = self.__class__.__name__
        dirs = ", ".join(f'"{d}"' for d in sorted(self.dirlist))
        return f"{mod}.{cname}(({dirs}, ), depth={self.depth})"

    def get_items(self) -> ty.Iterable[Leaf]:
        for directory in self.dirlist:
            dirfiles = fileutils.get_dirlist(
                directory, max_depth=self.depth, exclude=self._exclude_file
            )
            yield from map(construct_file_leaf, dirfiles)

    def should_sort_lexically(self) -> bool:
        return True

    def _exclude_file(self, filename: str) -> bool:
        return filename.startswith(".")

    def get_description(self) -> str:
        return _("Recursive source of %(dir)s, (%(levels)d levels)") % {
            "dir": self.name,
            "levels": self.depth,
        }

    def get_icon_name(self) -> str:
        return "folder-saved-search"

    def provides(self) -> ty.Iterator[ty.Type[Leaf]]:
        yield files.FileLeaf
        yield apps.AppLeaf



__kupfer_name__ = _("Instapaper")
__kupfer_sources__ = ()
__kupfer_actions__ = ("SaveToInstapaper",)
__description__ = _("Save url to Instapaper")
__version__ = "2020-11-15"
__author__ = "Peter Stuifzand <peter@p83.nl>"
import typing as ty

from kupfer import launch
from kupfer.obj import Action, TextLeaf, UrlLeaf

if ty.TYPE_CHECKING:
    from gettext import gettext as _


class SaveToInstapaper(Action):
    def __init__(self):
        Action.__init__(self, _("Save to Instapaper"))

    def activate(self, leaf, iobj=None, ctx=None):
        launch.show_url(f"https://www.instapaper.com/edit?url={leaf.object}")

    def item_types(self):
        yield UrlLeaf
        yield TextLeaf

    def get_description(self):
        return _("Save url to Instapaper")



"""
The main logic for executing constructed commands.

A command is normally a tuple of (object, action, indirect object).
Where, of course, the indirect object is often not needed (in this module we
then pass None in its stead).

This code was once a shining machine; While adding the "comma trick" and
support for "multiple dispatch" was easy in the rest of the program, it shed
its casualties here: While the main process is simple, we deal here with all
the exceptions that are, at the moment, tacked on.

The ActionExecutionContext (ACE) keeps track of its nested invocation, so that
we can catch the results of commands executed inside other commands. The
delegation mechanism allows a user of the ACE to indicate that the result of
the command should be passed on from the earlier (more nested) invocation.

Multiple dispatch is straightforward if the action implements the multiple
dispatch protocol. Is the protocol not implemented, the command is simply
"multiplied out": executed once for each object, or once for each combination
of object and indirect object.

With multiple command execution (and delegation), we must then process and
merge multiple return values.

This file is a part of the program kupfer, which is
released under GNU General Public License v3 (or any later version),
see the main program file, and COPYING for details.

TODO: delegation, run, combine_action_result_multiple need rethink / rebuild.
It's work quite well but is too complicated.
"""

from __future__ import annotations

import collections
import contextlib
import itertools
import sys
import typing as ty
from enum import IntEnum
from functools import partial

from gi.repository import GObject

from kupfer.core._support import get_leaf_members, is_multiple_leaf
from kupfer.obj import (
    Action,
    KupferObject,
    Leaf,
    OperationError,
    Source,
    SourceLeaf,
)
from kupfer.obj.compose import MultipleLeaf
from kupfer.obj.sources import MultiSource
from kupfer.support import pretty, task
from kupfer.ui import uiutils

if ty.TYPE_CHECKING:
    from gettext import gettext as _

    from kupfer.support.types import ExecInfo
    from kupfer.ui.uievents import GUIEnvironmentContext

__all__ = (
    "ActionExecutionContext",
    "ActionExecutionError",
    "ExecResult",
    "ExecutionToken",
    "default_action_execution_context",
)


class ExecResult(IntEnum):
    ZERO = 0  # when delegate
    NONE = 1
    OBJECT = 2
    SOURCE = 3
    ASYNC = 4
    REFRESH = 5

    @property
    def is_sync(self):
        return self in (ExecResult.OBJECT, ExecResult.SOURCE)


_MAX_LAST_RESULTS: ty.Final = 10

## CmdTuple keep information about command arguments (leaf, action and optional
## iobj)
CmdTuple = tuple[Leaf, Action, ty.Optional[Leaf]]

## Token keep information about (command id, and optional CmdTuple)
Token = tuple[int, ty.Optional[CmdTuple]]

## ActionResult is result (content) of action
ActionResult = ty.Union[Source, Leaf, task.Task, None]


# ActionResultRefresh is special leaf used as result of action that
# want refresh current leaves list.
ActionResultRefresh = Leaf(None, "ActionResultRefresh")


class ActionExecutionError(Exception):
    pass


class ExecutionToken:
    """A token object that an `Action` carries with it from `activate`.

    Must be used for access to current execution context, and to access the
    environment.
    """

    def __init__(
        self,
        aectx: ActionExecutionContext,
        async_token: Token,
        ui_ctx: GUIEnvironmentContext | None,
    ) -> None:
        """Initialize `ExecutionToken`.
        `aectx` is context of execution.
        `async_token` keeps data returned with late result
        `ui_ctx` is optional gui context
        """
        self._aectx: ActionExecutionContext = aectx
        self._token = async_token
        self._ui_ctx = ui_ctx

    def register_late_result(
        self, result_object: KupferObject, show: bool = True
    ) -> None:
        """Put execution result (`result_object`) into action execution
        context.
        If `show` it True - try to show result to the user.
        """
        self._aectx.register_late_result(
            self._token, result_object, show=show, ctxenv=self._ui_ctx
        )

    def register_late_error(
        self, exc_info: ExecInfo | BaseException | None = None
    ) -> None:
        """Put error (`exc_info`) from failed execution into action execution
        context."""
        self._aectx.register_late_error(self._token, exc_info)

    def delegated_run(
        self, obj: Leaf, action: Action, iobj: Leaf | None
    ) -> tuple[ExecResult, ty.Any]:
        """Activate action execution context - run it for `obj`, `action` and
        `iobj`."""
        return self._aectx.run(
            obj, action, iobj, delegate=True, ui_ctx=self._ui_ctx
        )

    @property
    def environment(self) -> GUIEnvironmentContext:
        """This is a property for the current environment,
        access env variables like this::

            ctx.environment.get_timestamp()

        Raises RuntimeError when not available.
        """
        if self._ui_ctx is not None:
            return self._ui_ctx

        raise RuntimeError("Environment Context not available")


class ActionExecutionContext(GObject.GObject, pretty.OutputMixin):  # type: ignore
    """The ActionExecutionContext (ACE) keeps track of its nested invocation,
    so that we can catch the results of commands executed inside other commands.
    The delegation mechanism allows a user of the ACE to indicate that the result
    of the command should be passed on from the earlier (more nested) invocation.

    ActionExecutionContext is singleton; to get it use
    `ActionExecutionContext.instance()` or `default_action_execution_context()`.

    command-result (result_type, result)
        Emitted when a command is carried out, with its resulting value
    """

    __gtype_name__ = "ActionExecutionContext"
    _instance: ActionExecutionContext | None = None

    @classmethod
    def instance(cls) -> ActionExecutionContext:
        if cls._instance is None:
            cls._instance = ActionExecutionContext()

        assert cls._instance
        return cls._instance

    def __init__(self) -> None:
        GObject.GObject.__init__(self)
        self._task_runner = task.TaskRunner(end_on_finish=False)
        self._nest_level = 0
        self._delegate = False
        self._command_counter = itertools.count()
        self.last_command_id = -1
        self.last_command: CmdTuple | None = None
        self._last_executed_command: CmdTuple | None = None
        self.last_results: collections.deque[ty.Any] = collections.deque(
            [], _MAX_LAST_RESULTS
        )

    @contextlib.contextmanager
    def _nesting(self):
        try:
            self._nest_level += 1
            self._delegate = False
            yield
        finally:
            self._nest_level -= 1

    @property
    def _is_nested(self) -> bool:
        return bool(self._nest_level)

    @contextlib.contextmanager
    def _error_conversion(self, cmdtuple: CmdTuple) -> ty.Any:
        try:
            yield
        except OperationError:
            self._do_error_conversion(cmdtuple, sys.exc_info())

    def _do_error_conversion(
        self, cmdtuple: CmdTuple, exc_info: ExecInfo
    ) -> None:
        if not self._operation_error(exc_info, cmdtuple):
            raise exc_info[1]  # type: ignore

        _etype, value, traceback = exc_info
        raise ActionExecutionError(value).with_traceback(traceback)

    def _get_async_token(self) -> Token:
        """Get an action execution for current execution

        Return a token for the currently active command execution.
        The token must be used for posting late results or late errors.
        """
        assert self.last_command_id is not None
        return (self.last_command_id, self._last_executed_command)

    def make_execution_token(
        self, ui_ctx: GUIEnvironmentContext | None
    ) -> ExecutionToken:
        """Return an ExecutionToken for @self and @ui_ctx."""
        return ExecutionToken(self, self._get_async_token(), ui_ctx)

    def _operation_error(
        self, exc_info: ExecInfo, cmdtuple: CmdTuple
    ) -> int | None:
        """Error when executing action - show notification to the user.
        Nested error are ignored."""

        if self._is_nested:
            return None

        _etype, value, _tb = exc_info
        _obj, action, _iobj = cmdtuple
        # TRANS: When an error occurs in an action to be carried out,
        # TRANS: then this is the heading of the error notification
        return uiutils.show_notification(
            _("Could not to carry out '%s'") % action,
            str(value),
            icon_name="kupfer",
        )

    def register_late_error(
        self, token: Token, exc_info: ExecInfo | BaseException | None = None
    ) -> None:
        """Register an error in exc_info. The error must be an OperationError.
        ActionExecutionError is raised.
        """
        if exc_info is None:
            exc_info = sys.exc_info()
        elif isinstance(exc_info, BaseException):
            exc_info = (type(exc_info), exc_info, None)

        _command_id, cmdtuple = token
        assert exc_info
        assert cmdtuple
        self._do_error_conversion(cmdtuple, exc_info)

    def register_late_result(
        self,
        token: Token,
        result: ty.Any,
        show: bool = True,
        ctxenv: GUIEnvironmentContext | None = None,
    ) -> None:
        """Register a late result.

        Result must be a Leaf (as in result object, not factory or async).

        If `show`, possibly display the result to the user (show leaf).
        Desktop notification is always displayed when there is any result.
        """
        self.output_debug("Late result", result, "for", token)
        assert token and token[1]
        command_id, (_ign1, action, _ign2) = token  # type: ignore
        if result is None:
            raise ActionExecutionError(f"Late result from {action} was None")

        assert isinstance(result, (Leaf, Source))
        description = res_name = str(result)
        if res_desc := result.get_description():
            description = f"{res_name} ({res_desc})"

        # If only registration was requested, remove the command id info
        if not show:
            command_id = -1

        result_type = _parse_late_action_result(action, result)

        self.output_debug(
            "late-command-result", command_id, result_type, result, ctxenv
        )

        if result_type == ExecResult.NONE:
            return

        if result_type != ExecResult.REFRESH:
            uiutils.show_notification(
                _('"%s" produced a result') % action, description
            )

        self.emit(
            "late-command-result", command_id, result_type, result, ctxenv
        )
        self._append_result(result_type, result)

    def _append_result(self, res_type: int, result: ty.Any) -> None:
        if res_type == ExecResult.OBJECT:
            self.last_results.append(result)

    def run(
        self,
        obj: Leaf,
        action: Action,
        iobj: Leaf | None,
        delegate: bool = False,
        ui_ctx: GUIEnvironmentContext | None = None,
    ) -> tuple[ExecResult, ty.Any]:
        """
        Activate the command (obj, action, iobj), where @iobj may be None

        Return a tuple (DESCRIPTION; RESULT)

        If a command carries out another command as part of its execution,
        and wishes to delegate to it, pass True for @delegate.
        """
        self.last_command_id = next(self._command_counter)
        self._last_executed_command = (obj, action, iobj)

        if not action or not obj:
            raise ActionExecutionError("Primary Object and Action required")

        if iobj is None and action.requires_object():
            raise ActionExecutionError(f"{action} requires indirect object")

        self.output_debug(obj, action, iobj, ui_ctx)

        # The execution token object for the current invocation
        execution_token = self.make_execution_token(ui_ctx)
        with self._error_conversion((obj, action, iobj)), self._nesting():
            ret = activate_action(execution_token, obj, action, iobj)

        # remember last command, but not delegated commands.
        if not delegate:
            self.last_command = self._last_executed_command

        # Delegated command execution was previously requested: we take
        # the result of the nested execution context
        if self._delegate:
            assert not ret or isinstance(ret, tuple)
            res, ret = ty.cast("tuple[ExecResult, ty.Any]", ret)
            return self._return_result(res, ret, ui_ctx)

        assert not ret or isinstance(ret, (Source, Leaf, task.Task))
        res = parse_action_result(action, ret)
        if res == ExecResult.ASYNC:
            # Register the task then "clear" the result
            self.output_debug("Registering async task", ret)
            assert isinstance(ret, task.Task)
            self._task_runner.add_task(ret)
            res, ret = ExecResult.NONE, None

        # Delegated command execution was requested: we pass
        # through the result of the action to the parent execution context
        if delegate and self._is_nested:
            self._delegate = True

        return self._return_result(res, ret, ui_ctx)

    def _return_result(
        self,
        res: ExecResult,
        ret: ty.Any,
        ui_ctx: GUIEnvironmentContext | None,
    ) -> tuple[ExecResult, ty.Any]:
        if not self._is_nested:
            self._append_result(res, ret)
            self.emit("command-result", res, ret, ui_ctx)

        return res, ret

    def combine_action_result_multiple(
        self,
        action: Action,
        retvals: ty.Iterable[tuple[ExecResult, ActionResult] | ActionResult],
    ) -> tuple[ExecResult, ActionResult] | ActionResult:
        """
        When delegate is False `retvals` is list of `ActionResult` and function
        return  `ActionResult`.

        When delegate it True  `retvals` is list of (ExecResult, ActionResult)
        and function return (ExecResult, ActionResult)

        """
        self.output_debug(
            "Combining", action, retvals, f"delegate={self._delegate}"
        )

        retvals_not_empty = filter(None, retvals)

        if self._delegate:
            return self._combine_action_result_multiple_delegate(
                action,
                ty.cast(
                    "list[tuple[ExecResult, ActionResult]]", retvals_not_empty
                ),
            )

        return self._combine_action_result_multiple_non_delegate(
            action, ty.cast("list[ActionResult]", retvals_not_empty)
        )

    def _combine_action_result_multiple_non_delegate(
        self,
        action: Action,
        retvals: ty.Iterable[ActionResult],
    ) -> ActionResult:
        values: list[ActionResult] = []
        res = ExecResult.NONE
        for ret in retvals:
            res_type = parse_action_result(action, ret)
            if res_type != ExecResult.NONE:
                values.append(ret)
                res = res_type

        return self._make_retvalue(res, values)

    def _combine_action_result_multiple_delegate(
        self,
        action: Action,
        retvals: list[tuple[ExecResult, ActionResult]],
    ) -> tuple[ExecResult, ActionResult]:
        # Re-parse result values
        resmap: dict[ExecResult, list[ActionResult]] = collections.defaultdict(
            list
        )
        for res_type, ret_obj in retvals:
            if res_type != ExecResult.NONE:
                resmap[res_type].append(ret_obj)

        # register tasks
        if tasks := resmap.pop(ExecResult.ASYNC, None):
            self._make_retvalue(ExecResult.ASYNC, tasks)

        if len(resmap) == 1:
            # Return the only of the Source or Object case
            key, values = resmap.popitem()
            return key, self._make_retvalue(key, values)

        if len(resmap) > 1:
            # Put the source in a leaf and return a multiple leaf
            source = self._make_retvalue(
                ExecResult.SOURCE, resmap[ExecResult.SOURCE]
            )
            assert isinstance(source, Source)
            objects = resmap[ExecResult.OBJECT]
            objects.append(SourceLeaf(source))
            return ExecResult.OBJECT, self._make_retvalue(
                ExecResult.OBJECT, objects
            )

        return ExecResult.NONE, None

    def _make_retvalue(
        self, res: ExecResult, values: list[ActionResult]
    ) -> ActionResult:
        """Construct a return value for type res.

        If Result is source or object construct MultpileSource/MultipleLeaf for
        many results or simple return first result.
        For async result register async task.
        """
        if res == ExecResult.SOURCE:
            return (
                values[0]
                if len(values) == 1
                else MultiSource(ty.cast("ty.Collection[Source]", values))
            )

        if res == ExecResult.OBJECT:
            return values[0] if len(values) == 1 else MultipleLeaf(values)

        if res == ExecResult.ASYNC:
            # Register all tasks now, and return None upwards
            for task_ in values:
                assert isinstance(task_, task.Task)
                self.output_debug("Registering async task", task_)
                self._task_runner.add_task(task_)

        return None


# Signature: Action result type, action result, gui_context
GObject.signal_new(
    "command-result",
    ActionExecutionContext,
    GObject.SignalFlags.RUN_LAST,
    GObject.TYPE_BOOLEAN,
    (GObject.TYPE_INT, GObject.TYPE_PYOBJECT, GObject.TYPE_PYOBJECT),
)

# Signature: Command ID, Action result type, action result, gui_context
GObject.signal_new(
    "late-command-result",
    ActionExecutionContext,
    GObject.SignalFlags.RUN_LAST,
    GObject.TYPE_BOOLEAN,
    (
        GObject.TYPE_INT,
        GObject.TYPE_INT,
        GObject.TYPE_PYOBJECT,
        GObject.TYPE_PYOBJECT,
    ),
)


## Get ActionExecutionContext instance
default_action_execution_context = ActionExecutionContext.instance


# pylint: disable=too-few-public-methods
@ty.runtime_checkable
class ActionActivateFunc(ty.Protocol):
    def __call__(
        self,
        leaf: Leaf,
        iobj: Leaf | None = None,
        ctx: ExecutionToken | None = None,
    ) -> ActionResult: ...


# pylint: disable=too-few-public-methods
@ty.runtime_checkable
class ActionActivateMultipleFunc(ty.Protocol):
    def __call__(
        self,
        leaf: ty.Iterable[Leaf],
        iobj: ty.Iterable[Leaf | None] | None = None,
        ctx: ExecutionToken | None = None,
    ) -> ActionResult: ...


def activate_action(
    context: ExecutionToken | None,
    obj: Leaf,
    action: Action,
    iobj: Leaf | None,
) -> tuple[ExecResult, ActionResult] | ActionResult:
    """Activate @action in simplest manner"""

    if not action.wants_context():
        context = None

    if not is_multiple_leaf(obj) and not is_multiple_leaf(iobj):
        return _activate_action_single(obj, action, iobj, context)

    return _activate_action_multiple(obj, action, iobj, context)


def _activate_action_single(
    obj: Leaf, action: Action, iobj: Leaf | None, ctx: ExecutionToken | None
) -> tuple[ExecResult, ActionResult] | ActionResult:
    func: ActionActivateFunc = action.activate
    if ctx:
        # set context to action.activate call
        func = partial(func, ctx=ctx)

    if action.requires_object():
        return func(obj, iobj)

    return func(obj)


def _activate_action_multiple(
    obj: Leaf, action: Action, iobj: Leaf | None, ctx: ExecutionToken | None
) -> tuple[ExecResult, ActionResult] | ActionResult | None:
    objs = get_leaf_members(obj)

    if not hasattr(action, "activate_multiple"):
        iobjs = (None,) if iobj is None else get_leaf_members(iobj)
        return _activate_action_multiple_multiplied(objs, action, iobjs, ctx)

    func: ActionActivateMultipleFunc = action.activate_multiple
    if ctx:
        func = partial(func, ctx=ctx)

    if action.requires_object():
        iobjs = (None,) if iobj is None else get_leaf_members(iobj)
        return func(objs, iobjs)

    return func(objs)


def _activate_action_multiple_multiplied(
    objs: ty.Iterable[Leaf],
    action: Action,
    iobjs: ty.Iterable[Leaf | None],
    ctx: ExecutionToken | None,
) -> tuple[ExecResult, ActionResult] | ActionResult | None:
    """
    Multiple dispatch by "multiplied" invocation of the simple activation

    When action is delegated return (ExecResult, ActionResult), otherwise
    return ActionResult

    """
    rets = [
        _activate_action_single(leaf, action, item, ctx)
        for leaf in objs
        for item in iobjs
    ]

    actx = default_action_execution_context()
    return actx.combine_action_result_multiple(action, rets)


def parse_action_result(action: Action, ret: ActionResult) -> ExecResult:
    """Return result type for @action and return value @ret"""
    if ret is ActionResultRefresh:
        return ExecResult.REFRESH

    if not ret or (hasattr(ret, "is_valid") and not ret.is_valid()):
        return ExecResult.NONE

    # handle actions returning "new contexts"
    res = ExecResult.NONE
    if action.is_factory():
        res = ExecResult.SOURCE

    if action.has_result():
        res = ExecResult.OBJECT
    elif action.is_async():
        res = ExecResult.ASYNC

    return res


def _parse_late_action_result(action: Action, ret: ty.Any) -> int:
    # Late result is assumed to be a Leaf (Object) result
    # by default for backward compat.
    #
    # It is also allowed to be a Source

    if ret is ActionResultRefresh:
        return ExecResult.REFRESH

    if not ret or (hasattr(ret, "is_valid") and not ret.is_valid()):
        return ExecResult.NONE

    if isinstance(ret, Source):
        return ExecResult.SOURCE

    return ExecResult.OBJECT



"""
Kupfer's Hosts API

Main definition and *constructor* classes.

This file is a part of the program kupfer, which is
released under GNU General Public License v3 (or any later version),
see the main program file, and COPYING for details.
"""

from __future__ import annotations

import typing as ty

from kupfer.obj.grouping import GroupingLeaf, Slots

__author__ = (
    "Ulrik Sverdrup <ulrik.sverdrup@gmail.com>, "
    "Karol Będkowski <karol.bedkowsk+gh@gmail.com>"
)

__all__ = (
    "HostLeaf",
    "HostServiceLeaf",
)

HOST_NAME_KEY = "HOST_NAME"
HOST_ADDRESS_KEY = "HOST_ADDRESS"
HOST_SERVICE_NAME_KEY = "HOST_SERVICE_NAME"
HOST_SERVICE_PORT_KEY = "HOST_SERVICE_PORT"
HOST_SERVICE_USER_KEY = "HOST_SERVICE_USER"
HOST_SERVICE_PASS_KEY = "HOST_SERVICE_PASS"
HOST_SERVICE_REMOTE_PATH_KEY = "HOST_SERVICE_REMOTE_PATH"


class HostLeaf(GroupingLeaf):
    grouping_slots = (HOST_NAME_KEY, HOST_ADDRESS_KEY)

    def __init__(self, obj: dict[str, ty.Any], name: str) -> None:
        super().__init__(obj, name)
        if service := obj.get(HOST_SERVICE_NAME_KEY):
            if hostname := obj.get(HOST_NAME_KEY):
                self.kupfer_add_alias(f"{service}://{hostname}")

            if address := obj.get(HOST_ADDRESS_KEY):
                self.kupfer_add_alias(f"{service}://{address}")

    def get_icon_name(self) -> str:
        return "computer"

    def get_text_representation(self):
        obj = self.object
        if service := obj.get(HOST_SERVICE_NAME_KEY):
            if hostname := obj.get(HOST_NAME_KEY):
                return f"{service}://{hostname}"

            if address := obj.get(HOST_ADDRESS_KEY):
                return f"{service}://{address}"

        return self.name


class HostServiceLeaf(HostLeaf):
    """Leaf dedicated for well known services like ftp, ssh, vnc"""

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        name: str,
        address: str,
        service: str,
        description: str,
        port: str | None = None,
        user: str | None = None,
        password: str | None = None,
        slots: Slots = None,
    ) -> None:
        _slots = {
            HOST_NAME_KEY: name,
            HOST_ADDRESS_KEY: address,
            HOST_SERVICE_NAME_KEY: service,
            HOST_SERVICE_PORT_KEY: port,
            HOST_SERVICE_USER_KEY: user,
            HOST_SERVICE_PASS_KEY: password,
        }
        if slots:
            _slots.update(slots)

        HostLeaf.__init__(self, _slots, name or address)
        self._description = description

    def get_description(self) -> str:
        return self._description



# Copyright (C) 2009  Ulrik Sverdrup <ulrik.sverdrup@gmail.com>
#               2008  Christian Hergert <chris@dronelabs.com>
#               2007  Chris Halse Rogers, DR Colkitt
#                     David Siegel, James Walker
#                     Jason Smith, Miguel de Icaza
#                     Rick Harding, Thomsen Anders
#                     Volker Braun, Jonathon Anderson
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
This module provides relevance matching and formatting of related strings
based on the relevance.  It originates in Gnome-Do.

 * Python port by Christian Hergert

 * Module updated by Ulrik Sverdrup to clean up and dramatically speed up
   the code, by using more pythonic constructs as well as doing less work.

Compatibility: Python 3
"""

from __future__ import annotations

import typing as ty

__all__ = ("format_common_substrings", "score", "score_single")

FormatCleanCB = ty.Callable[[str], str]
FormatMatchCB = ty.Callable[[str], str]


def _default_formatter(x: str) -> str:
    return x


def format_common_substrings(
    string: str,
    query: str,
    format_clean: FormatCleanCB | None = None,
    format_match: FormatMatchCB | None = None,
) -> str:
    """Creates a new string highlighting matching substrings.
    Returns: a formatted string

    >>> format_common_substrings('hi there dude', 'hidude',
    ...                        format_match=lambda m: "<b>%s</b>" % m)
    '<b>hi</b> there <b>dude</b>'

    >>> format_common_substrings('parallelism', 'lsm', format_match=str.upper)
    'paralleLiSM'
    """
    format_clean = format_clean or _default_formatter

    def _format(x: str) -> str:
        return x and format_clean(x)

    if not query:
        return _format(string)

    lowerstr = string.lower()

    # find overall range of match
    first, last = _find_best_match(lowerstr, query)

    if first == -1:
        return _format(string)

    # find longest perfect match, put in slc
    for slc in range(len(query), 0, -1):
        if query[:slc] == lowerstr[first : first + slc]:
            break

    nextkey = query[slc:]
    head = string[:first]
    match = string[first : first + slc]
    matchtail = string[first + slc : last]
    tail = string[last:]

    format_match = format_match or _default_formatter

    return "".join(
        (
            _format(head),
            format_match(match),
            format_common_substrings(
                matchtail, nextkey, format_clean, format_match
            ),
            _format(tail),
        )
    )


def score_single(string: str, query: str) -> float:
    """Find the shortest possible substring that matches the query
     and get the ration of their lengths for a base score.

    string: text body to score
    query: A single character

    This is a single character approximation to `score`.

    >>> round(score_single('terminal', 't'), 6)
    0.973125
    >>> round(score_single('terminal', 'e'), 6)
    0.903125
    >>> round(score_single('terminal', 'a'), 6)
    0.903125
    >>> round(score_single('t', 't'), 6)
    0.995
    """

    string = string.lower()

    first = string.find(query)
    if first == -1:
        return 0.0

    if first == 0:
        return 0.97 + 0.025 / len(string)

    return 0.9 + 0.025 / len(string)


def score(string: str, query: str) -> float:
    """A relevancy score for the string ranging from 0 to 1.

    @string: a string to be scored
    @query: a string query to score against

    `string' is treated case-insensitively while `query' is interpreted
    literally, including case and whitespace.

    Returns: a float between 0 and 1

    >>> round(score('terminal', 'trml'), 6)
    0.735099
    >>> round(score('terminal', 'term'), 6)
    0.992303
    >>> print(score('terminal', 'try'))
    0.0
    >>> print(score('terminal', ''))
    1.0
    >>> round(score('terminal', 't'), 6)
    0.98949
    >>> round(score('terminal', 'e'), 6)
    0.918438
    """
    if not query:
        return 1.0

    string = string.lower()

    # Find the shortest possible substring that matches the query
    # and get the ration of their lengths for a base score
    first, last = _find_best_match(string, query)
    if first == -1:
        return 0.0

    query_len = len(query)
    strscore: float = query_len / (last - first)

    # Now we weight by string length so shorter strings are better
    strscore *= 0.7 + query_len / len(string) * 0.3

    # Bonus points if the characters start words
    bad = 1
    first_count = 0
    for i in range(first, last - 1):
        if string[i] in " -.([_":
            if string[i + 1] in query:
                first_count += 1
            else:
                bad += 1

    # A first character match counts extra
    if query[0] == string[0]:
        first_count += 2

    # The longer the acronym, the better it scores
    good = first_count * first_count * 4

    # Better yet if the match itself started there
    if first == 0:
        good += 2

    # Super duper bonus if it is a perfect match
    if query == string:
        good += last * 2 + 4

    strscore = (strscore + 3 * good / (good + bad)) / 4

    # This fix makes sure that perfect matches always rank higher
    # than split matches.  Perfect matches get the .9 - 1.0 range
    # everything else lower

    if last - first == query_len:
        return 0.9 + 0.1 * strscore

    return 0.9 * strscore


def _find_best_match(string: str, query: str) -> tuple[int, int]:
    """Finds the shortest substring of @s that contains all characters of query
    in order.

    @string: a string to search
    @query: a string query to search for

    Returns: a two-item tuple containing the start and end indices of
             the match.  No match returns (-1,-1).

    >>> _find_best_match('terminal', 'trml')
    (0, 8)
    >>> _find_best_match('total told', 'tl')
    (2, 5)
    >>> _find_best_match('terminal', 'yl')
    (-1, -1)
    """
    # Find the last instance of the last character of the query
    # since we never need to search beyond that
    last_char = string.rfind(query[-1])

    # No instance of the character?
    if last_char == -1:
        return -1, -1

    # Loop through each instance of the first character in query
    index = string.find(query[0])
    best_match_start = -1
    best_match_end = -1
    query_len = len(query)
    last_index = last_char - query_len + 1
    while 0 <= index <= last_index:
        # See if we can fit the whole query in the tail
        # We know the first char matches, so we dont check it.
        cur = index + 1
        qcur = 1
        while qcur < query_len:
            # find where in the string the next query character is
            # if not found, we are done
            cur = string.find(query[qcur], cur, last_char + 1)
            if cur == -1:
                return best_match_start, best_match_end

            cur += 1
            qcur += 1

        # take match if it is shorter
        # if perfect match, we are done
        if best_match_start == -1 or (cur - index) < (
            best_match_end - best_match_start
        ):
            best_match_start = index
            best_match_end = cur
            if cur - index == query_len:
                break

        index = string.find(query[0], index + 1)

    return best_match_start, best_match_end


if __name__ == "__main__":
    import doctest

    doctest.testmod()
